/*
 *  Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */

/*
 *  Written by Anatoly P. Pinchuk pap@namesys.botik.ru
 *  Programm System Institute
 *  Pereslavl-Zalessky Russia
 */

/*
 *  This file contains functions dealing with S+tree
 *
 * comp_keys
 * comp_short_keys
 * bin_search
 * get_lkey
 * get_rkey
 * key_in_buffer
 * decrement_bcount
 * decrement_counters_in_path
 * pathrelse
 * search_by_key
 * search_for_position_by_key
 * comp_items
 * prepare_for_delete_or_cut
 * calc_deleted_bytes_number
 * init_tb_struct
 * reiserfs_delete_item
 * reiserfs_delete_object
 * indirect_to_direct
 * maybe_indirect_to_direct
 * reiserfs_cut_from_item
 * reiserfs_truncate_file
 * reiserfs_paste_into_item
 * reiserfs_insert_item
 * get_buffer_by_range
 * get_buffers_from_range
 */
#ifdef __KERNEL__

#include <linux/sched.h>
#include <linux/string.h>
#include <linux/reiserfs_fs.h>

#else

#include "nokernel.h"

#endif


/* Does the buffer contain a disk block which is in the tree. */
inline int B_IS_IN_TREE(
			struct buffer_head * p_s_bh
			) {

#ifdef REISERFS_CHECK
  if ( B_BLK_HEAD(p_s_bh)->blk_level > MAX_HEIGHT ) {
    print_buffer_head (p_s_bh, "");
    reiserfs_panic(0, "PAP-1010: B_IS_IN_TREE: formatted node has level %d. Must be less than %d",
		   B_BLK_HEAD(p_s_bh)->blk_level, MAX_HEIGHT + 1);
  }
#endif

  return ( B_BLK_HEAD(p_s_bh)->blk_level != FREE_LEVEL );
}


inline void copy_key (void * to, void * from)
{
  memcpy (to, from, KEY_SIZE);
}


inline void copy_object_locale (struct key * to, struct key * from)
{
  to->group_locale = from->group_locale;
  to->name_hash = from->name_hash;
  to->file_id = from->file_id;
}


inline void copy_item_head(void * p_v_to, void * p_v_from)
{
  memcpy (p_v_to, p_v_from, IH_SIZE);
}



/*
 Compare keys using all key fields.
 Returns:  -1 if key1 < key2
            0 if key1 = key2
            1 if key1 > key2
*/
#define compare_fields(f1,f2) \
if (f1 < f2)\
return -1;\
if (f1 > f2)\
return 1;

inline int comp_keys (struct key * k1, struct key * k2)
{
  compare_fields (k1->group_locale.major, k2->group_locale.major);
  compare_fields (k1->group_locale.minor, k2->group_locale.minor);
  compare_fields (k1->name_hash, k2->name_hash);
  compare_fields (k1->file_id, k2->file_id);
  compare_fields (k1->k_offset, k2->k_offset);
  return 0;
}


/* return 0 if keys are of the same object, 1 otherwise */
inline int items_of_different_files (struct key * key1, struct key * key2)
{
  return (key1->group_locale.major  == key2->group_locale.major && 
	  key1->name_hash == key2->name_hash &&
	  key1->file_id == key2->file_id) ? 0 : 1;
}


/* return 0 if keys are of the same item types, 1 otherwise */
inline int keys_of_different_item_types (struct key * key1, struct key * key2)
{
  if ((is_key_of_stat_data (key1) && is_key_of_stat_data (key2)) ||
      (is_key_of_direct_item (key1) && is_key_of_direct_item (key2)) ||
      (is_key_of_indirect_item (key1) && is_key_of_indirect_item (key2)) ||
      (is_key_of_dir_entry (key1) && is_key_of_dir_entry (key2)))
    return 0;
  return 1;
}


#define comp_short_keys items_of_different_files



/*
 Compare keys using REISERFS_SHORT_KEY_LEN fields.
 Returns:  -1 if key1 < key2
            0 if key1 = key2
            1 if key1 > key2
 Used to check whether keys are of the same object
*/
/*
this is changed with "items_of_different_files"
inline int  comp_short_keys (struct key * k1, struct key * k2)
{
  int n_key_length = REISERFS_SHORT_KEY_LEN;
  __u32 * p_s_key1 = (__u32 *)k1;
  __u32 * p_s_key2 = (__u32 *)k2;

  for( ; n_key_length--; ++p_s_key1, ++p_s_key2 ) {
    if ( *p_s_key1 < *p_s_key2 )
      return -1;
    if ( *p_s_key1 > *p_s_key2 )
      return 1;
  }

  return 0;
}
*/

/**************************************************************************
 *  Binary search toolkit function                                        *
 *  Search for an item in the array by the item key                       *
 *  Returns:    1 if found,  0 if not found;                              *
 *        *p_n_pos = number of the searched element if found, else the    *
 *        number of the first element that is larger than p_v_key.        *
 **************************************************************************/
/* For those not familiar with binary search: n_lbound is the leftmost item that it could be,
   n_rbound the rightmost item that it could be.  We examine the item halfway between n_lbound and
   n_rbound, and that tells us either that we can increase n_lbound, or decrease n_rbound, or that
   we have found it, or if n_lbound <= n_rbound that there are no possible items, and we have not
   found it. With each examination we cut the number of possible items it could be by one more than
   half rounded down, or we find it. */
inline	int bin_search (
              struct key    * p_s_key,    /* Key to search for.                   */
	      void    * p_v_base,   /* First item in the array.             */
	      int       p_n_num,    /* Number of items in the array.        */
	      int       p_n_width,  /* Item size in the array.  searched. Lest the reader be
				       confused, note that this is crafted as a general function,
				       and when it is applied specifically to the array of item
				       headers in a node, p_n_width is actually the item header size
				       not the item size.  */
	      int     * p_n_pos     /* Number of the searched for element. */
            ) {
  int   n_rbound, n_lbound, n_j;

  for ( n_j = ((n_rbound = p_n_num - 1) + (n_lbound = 0))/2; n_lbound <= n_rbound; n_j = (n_rbound + n_lbound)/2 )
    switch( comp_keys((struct key *)((char * )p_v_base + n_j * p_n_width), p_s_key) )  {
    case -1: n_lbound = n_j + 1; continue;
    case  1: n_rbound = n_j - 1; continue;
    case  0: *p_n_pos = n_j;     return ITEM_FOUND; /* Key found in the array.  */
    }

  /* bin_search did not find given key, it returns position of key,
     that is minimal and greater than the given one. */
  *p_n_pos = n_lbound;
  return ITEM_NOT_FOUND;
}

#ifdef REISERFS_CHECK
extern struct tree_balance * cur_tb;
#endif



/* Minimal possible key. It is never in the tree. */
struct key  MIN_KEY = {{0,0}, 0, 0, 0};

/* Maximal possible key. It is never in the tree. */
struct key  MAX_KEY = {{3,0x3fffffff}, 0xffffffff, 0xffffffff, 0xffffffff};


/* Get delimiting key of the buffer by looking for it in the buffers in the path, starting from the bottom
   of the path, and going upwards.  We must check the path's validity at each step.  If the key is not in
   the path, there is no delimiting key in the tree (buffer is first or last buffer in tree), and in this
   case we return a special key, either MIN_KEY or MAX_KEY. */
inline	struct  key * get_lkey  (
	                struct path         * p_s_chk_path,
                        struct super_block  * p_s_sb
                      ) {
  int                   n_position, n_path_offset = p_s_chk_path->path_length;
  struct buffer_head  * p_s_parent;
  
#ifdef REISERFS_CHECK
  if ( n_path_offset < FIRST_PATH_ELEMENT_OFFSET )
    reiserfs_panic(p_s_sb,"PAP-5010: get_lkey: illegal offset in the path");
#endif

  /* While not higher in path than first element. */
  while ( n_path_offset-- > FIRST_PATH_ELEMENT_OFFSET ) {

#ifdef REISERFS_CHECK
    if ( ! buffer_uptodate(PATH_OFFSET_PBUFFER(p_s_chk_path, n_path_offset)) )
      reiserfs_panic(p_s_sb, "PAP-5020: get_lkey: parent is not uptodate");
#endif

    /* Parent at the path is not in the tree now. */
    if ( ! B_IS_IN_TREE(p_s_parent = PATH_OFFSET_PBUFFER(p_s_chk_path, n_path_offset)) )
      return &MAX_KEY;
    /* Check whether position in the parent is correct. */
    if ( (n_position = PATH_OFFSET_POSITION(p_s_chk_path, n_path_offset)) > B_NR_ITEMS(p_s_parent) )
       return &MAX_KEY;
    /* Check whether parent at the path really points to the child. */
    if ( B_N_CHILD_NUM(p_s_parent, n_position) !=
	 PATH_OFFSET_PBUFFER(p_s_chk_path, n_path_offset + 1)->b_blocknr )
      return &MAX_KEY;
    /* Return delimiting key if position in the parent is not equal to zero. */
    if ( n_position )
      return B_N_PDELIM_KEY(p_s_parent, n_position - 1);
  }
  /* Return MIN_KEY if we are in the root of the buffer tree. */
  if ( PATH_OFFSET_PBUFFER(p_s_chk_path, FIRST_PATH_ELEMENT_OFFSET)->b_blocknr ==
                                      p_s_sb->u.reiserfs_sb.s_rs->s_root_block )
    return &MIN_KEY;
  return  &MAX_KEY;
}


/* Get delimiting key of the buffer at the path and its right neighbor. */
inline	struct  key * get_rkey  (
	                struct path         * p_s_chk_path,
                        struct super_block  * p_s_sb
                      ) {
  int                   n_position,
    			n_path_offset = p_s_chk_path->path_length;
  struct buffer_head  * p_s_parent;

#ifdef REISERFS_CHECK
  if ( n_path_offset < FIRST_PATH_ELEMENT_OFFSET )
    reiserfs_panic(p_s_sb,"PAP-5030: get_rkey: illegal offset in the path");
#endif

  while ( n_path_offset-- > FIRST_PATH_ELEMENT_OFFSET ) {

#ifdef REISERFS_CHECK
    if ( ! buffer_uptodate(PATH_OFFSET_PBUFFER(p_s_chk_path, n_path_offset)) )
      reiserfs_panic(p_s_sb, "PAP-5040: get_rkey: parent is not uptodate");
#endif

    /* Parent at the path is not in the tree now. */
    if ( ! B_IS_IN_TREE(p_s_parent = PATH_OFFSET_PBUFFER(p_s_chk_path, n_path_offset)) )
      return &MIN_KEY;
    /* Check whether position in the parrent is correct. */
    if ( (n_position = PATH_OFFSET_POSITION(p_s_chk_path, n_path_offset)) > B_NR_ITEMS(p_s_parent) )
      return &MIN_KEY;
    /* Check whether parent at the path really points to the child. */
    if ( B_N_CHILD_NUM(p_s_parent, n_position) !=
                                        PATH_OFFSET_PBUFFER(p_s_chk_path, n_path_offset + 1)->b_blocknr )
      return &MIN_KEY;
    /* Return delimiting key if position in the parent is not the last one. */
    if ( n_position != B_NR_ITEMS(p_s_parent) )
      return B_N_PDELIM_KEY(p_s_parent, n_position);
  }
  /* Return MAX_KEY if we are in the root of the buffer tree. */
  if ( PATH_OFFSET_PBUFFER(p_s_chk_path, FIRST_PATH_ELEMENT_OFFSET)->b_blocknr ==
       p_s_sb->u.reiserfs_sb.s_rs->s_root_block )
    return &MAX_KEY;
  return  &MIN_KEY;
}


/* Check whether a key is contained in the tree rooted from a buffer at a path. */
/* This works by looking at the left and right delimiting keys for the buffer in the last path_element in
   the path.  These delimiting keys are stored at least one level above that buffer in the tree. If the
   buffer is the first or last node in the tree order then one of the delimiting keys may be absent, and in
   this case get_lkey and get_rkey return a special key which is MIN_KEY or MAX_KEY. */
static  inline  int key_in_buffer (
                      struct path         * p_s_chk_path, /* Path which should be checked.  */
                      struct key          * p_s_key,      /* Key which should be checked.   */
                      struct super_block  * p_s_sb        /* Super block pointer.           */
		      ) {

#ifdef REISERFS_CHECK
  if ( ! p_s_key || p_s_chk_path->path_length < FIRST_PATH_ELEMENT_OFFSET ||
       p_s_chk_path->path_length > MAX_HEIGHT )
    reiserfs_panic(p_s_sb, "PAP-5050: key_in_buffer:  pointer to the key(%p) is NULL or illegal path length(%d)",
		   p_s_key, p_s_chk_path->path_length);
  
  if ( PATH_PLAST_BUFFER(p_s_chk_path)->b_dev == NODEV )
    reiserfs_panic(p_s_sb, "PAP-5060: key_in_buffer: device must not be NODEV");
#endif

  if ( comp_keys(get_lkey(p_s_chk_path, p_s_sb), p_s_key) == 1 )
    return 0;
  if ( comp_keys(p_s_key, get_rkey(p_s_chk_path, p_s_sb)) != -1 )
    return 0;
  return 1;
}


inline void decrement_bcount(
              struct buffer_head  * p_s_bh
            ) { 
  if ( p_s_bh ) {
    if ( p_s_bh->b_count ) {
      p_s_bh->b_count--;
      return;
    }

#ifdef REISERFS_CHECK
    print_buffer_head(p_s_bh, "decrement_bcount");
    reiserfs_panic(NULL, "PAP-5070: decrement_bcount: b_count equals zero");
#endif

    printk("decrement_bcount: trying to free free buffer\n");
  } 
}


void init_path (struct path * path)
{
  path->path_length = ILLEGAL_PATH_ELEMENT_OFFSET;
}


/* Decrement b_count field of the all buffers in the path. */
void decrement_counters_in_path (
              struct path * p_s_search_path
            ) {
  int n_path_offset = p_s_search_path->path_length;

#ifdef REISERFS_CHECK
  if ( n_path_offset < ILLEGAL_PATH_ELEMENT_OFFSET ||
       n_path_offset > EXTENDED_MAX_HEIGHT - 1 )
    reiserfs_panic(NULL, "PAP-5080: decrement_counters_in_path: illegal path offset of %d", n_path_offset);
#endif

  while ( n_path_offset > ILLEGAL_PATH_ELEMENT_OFFSET )
    decrement_bcount(PATH_OFFSET_PBUFFER(p_s_search_path, n_path_offset--));

  init_path (p_s_search_path);
}




/* Release all buffers in the path. */
void  pathrelse (
        struct path * p_s_search_path
      ) {
  int n_path_offset = p_s_search_path->path_length;

#ifdef REISERFS_CHECK
  if ( n_path_offset < ILLEGAL_PATH_ELEMENT_OFFSET )
    reiserfs_panic(NULL, "PAP-5090: pathrelse: illegal path offset");
#endif
  
  while ( n_path_offset > ILLEGAL_PATH_ELEMENT_OFFSET ) 
    brelse(PATH_OFFSET_PBUFFER(p_s_search_path, n_path_offset--));

  init_path (p_s_search_path);
}

/*
static int search_by_key_reada (struct super_block * s, int blocknr)
{
  struct buffer_head * bh;
  int repeat;
  
  repeat = CARRY_ON;
  if (blocknr == 0)
    return CARRY_ON;

  bh = reiserfs_getblk (s->s_dev, blocknr, s->s_blocksize, &repeat);
  
  if (!buffer_uptodate (bh)) {
    ll_rw_block (READA, 1, &bh);
    repeat = SCHEDULE_OCCURRED;
  }
  bh->b_count --;
  return repeat;
}
*/

/**************************************************************************
 * Algorithm   SearchByKey                                                *
 *             look for item in the Disk S+Tree by its key                *
 * Input:  p_s_sb   -  super block                                        *
 *         p_s_key  - pointer to the key to search                        *
 * returns: ITEM_FOUND or ITEM_NOT_FOUND                                  *
 *         p_s_search_path - path from the root to the needed leaf        *
 **************************************************************************/

/* This function fills up the path from the root to the leaf as it descends the tree looking for the
   key.  It uses reiserfs_bread to try to find buffers in the cache given their block number.  If it
   does not find them in the cache it reads them from disk.  For each node search_by_key finds using
   reiserfs_bread it then uses bin_search to look through that node.  bin_search will find the
   position of the block_number of the next node if it is looking through an internal node.  If it
   is looking through a leaf node bin_search will find the position of the item which has key either
   equal to given key, or which is the maximal key less than the given key.  search_by_key returns a
   path that must be checked for the correctness of the top of the path but need not be checked for
   the correctness of the bottom of the path */
int search_by_key(
                  struct super_block  * p_s_sb,         /* Super block.                           */
                  struct key          * p_s_key,        /* Key to search.                         */
                  struct path         * p_s_search_path,/* This structure was allocated and initialized by
                                                           the calling function. It is filled up by this
                                                           function.  */
                  int                 * p_n_repeat,     /* Whether schedule occured. */
                  int                   n_stop_level,   /* How far down the tree to search.*/
                  int                   n_bread_par     /* Whether to search even if it requires disk I/O, this is
                                                           either READ_BLOCKS or DONT_READ_BLOCKS or 0. Hans doesn't
                                                           know what 0 means, it seems to evaluate to DONT_READ_BLOCKS,
                                                           but it is bad style to not use the macro.... there is a
                                                           #define of search by key with no explanation that can allow
                                                           it to happen.... */
                  ) {
    kdev_t                      n_dev           = p_s_sb->s_dev;
    int                         n_repeat,
                                n_block_number  = p_s_sb->u.reiserfs_sb.s_rs->s_root_block,
                                n_block_size    = p_s_sb->s_blocksize;
    struct buffer_head  *       p_s_bh;
    struct path_element *       p_s_last_element;
    int				n_node_level, n_retval;
/*    int 			right_neighbor_of_leaf_node;*/

#ifdef REISERFS_CHECK
    int n_repeat_counter = 0;
#endif

    /* As we add each node to a path we increase its count.  This means that we must be careful to
       release all nodes in a path before we either discard the path struct or re-use the path
       struct, as we do here. */

    decrement_counters_in_path(p_s_search_path);

    *p_n_repeat = CARRY_ON;
/*    right_neighbor_of_leaf_node = 0;*/

    /* With each iteration of this loop we search through the items in the current node, and
       calculate the next current node(next path element) for the next iteration of this loop.. */
    while ( 1 ) {

#ifdef REISERFS_CHECK
      if ( !(++n_repeat_counter % 50000) )
	reiserfs_warning ("PAP-5100: search_by_key(pid %u): there were %d attemps from the tree root looking for key %k\n",
			  n_repeat_counter, p_s_key);
#endif

      /* prep path to have another element added to it. */
      p_s_last_element = PATH_OFFSET_PELEMENT(p_s_search_path, ++p_s_search_path->path_length);
      n_repeat = CARRY_ON;

      if ( n_bread_par == READ_BLOCKS ) { 
	/* schedule read of right neighbor */
/*	n_repeat |= search_by_key_reada (p_s_sb, right_neighbor_of_leaf_node);*/

	/* Read the next tree node, and set the last element in the path to have a pointer to it. */
	if ( ! (p_s_bh = p_s_last_element->pe_buffer =
		reiserfs_bread(n_dev, n_block_number, n_block_size, &n_repeat)) ) {
	  pathrelse(p_s_search_path);
	  *p_n_repeat |= n_repeat;
	  return ITEM_NOT_FOUND;	/* IO error */
	}
      }
      else { /* We are looking for the next tree node in cache. */
	p_s_bh = p_s_last_element->pe_buffer = reiserfs_getblk(n_dev, n_block_number, n_block_size, &n_repeat);
      }

      *p_n_repeat |= n_repeat;

      /* It is possible that schedule occured. We must check whether the key to search is still in
	 the tree rooted from the current buffer. If not then repeat search from the root. */
      if ( n_repeat != CARRY_ON && ((buffer_uptodate (p_s_bh) && !B_IS_IN_TREE (p_s_bh)) ||
				    (! key_in_buffer(p_s_search_path, p_s_key, p_s_sb))) ) { /* in fact this checks whether path is correct */
	decrement_counters_in_path(p_s_search_path);
	
	/* Get the root block number so that we can repeat the search starting from the root. */
	n_block_number  = p_s_sb->u.reiserfs_sb.s_rs->s_root_block;

/*	right_neighbor_of_leaf_node = 0;*/

	/* repeat search from the root */
	continue;
      }
      
#ifdef REISERFS_CHECK

      if ( ! key_in_buffer(p_s_search_path, p_s_key, p_s_sb) )
	reiserfs_panic(p_s_sb, "PAP-5130: search_by_key: key is not in the buffer");
      if ( cur_tb )
	reiserfs_panic(p_s_sb, "PAP-5140: search_by_key: schedule occurred in do_balance!");

#endif

      if ( ! buffer_uptodate(p_s_bh) ) {

#ifdef REISERFS_CHECK
	if ( n_bread_par != DONT_READ_BLOCKS )
	  reiserfs_panic(p_s_sb, "PAP-5150: search_by_key: buffer is not uptodate in case of READ_BLOCKS");
#endif

	return ITEM_NOT_FOUND; /* We can not continue search in the cache. */
      }

      /* ok, we have acquired next formatted node in the tree */
      n_node_level = B_BLK_HEAD(p_s_bh)->blk_level;

#ifdef REISERFS_CHECK

      if (n_node_level < n_stop_level)
	reiserfs_panic (p_s_sb, "vs-5152: search_by_key: tree level is less than stop level (%d)",
			n_node_level, n_stop_level);

#endif

      n_retval = bin_search (p_s_key, B_N_PITEM_HEAD(p_s_bh, 0), B_NR_ITEMS(p_s_bh),
		       ( n_node_level == DISK_LEAF_NODE_LEVEL ) ? IH_SIZE : KEY_SIZE, &(p_s_last_element->pe_position));
      if (n_node_level == n_stop_level)
	return n_retval;

      /* we are not in the stop level */
      if (n_retval == ITEM_FOUND)
	/* item has been found, so we choose the pointer which is to the right of the found one */
	p_s_last_element->pe_position++;
      /* if item was not found we choose the position which is to the left of the found item. This
	 requires no code, bin_search did it already.*/


      /* So we have chosen a position in the current node which is an
	 internal node.  Now we calculate child block number by position in the node. */
      n_block_number = B_N_CHILD_NUM(p_s_bh, p_s_last_element->pe_position);

      /* if we are going to read leaf node, then calculate its right neighbor if possible */
/*      if (n_node_level == DISK_LEAF_NODE_LEVEL + 1 && p_s_last_element->pe_position < B_NR_ITEMS (p_s_bh))
	right_neighbor_of_leaf_node = B_N_CHILD_NUM(p_s_bh, p_s_last_element->pe_position + 1);*/
    }
}



/* Form the path to an item and position in this item which contains file byte defined by p_s_key. If there
    is no such item corresponding to the key, we point the path to the item with maximal key less than
    p_s_key, and *p_n_pos_in_item is set to one past the last entry/byte in the item.  If searching for
    entry in a directory item, and it is not found, *p_n_pos_in_item is set to one entry more than the entry with
    maximal key which is less than the sought key.  

    Note that if there is no entry in this same node which is one more, then we point to an imaginary entry.

    for direct items, the position is in units of bytes, for
    indirect items the position is in units of blocknr entries, for directory items the position is in units
    of directory entries.  */
int search_for_position_by_key (
      struct super_block  * p_s_sb,         /* Pointer to the super block.          */
      struct key          * p_s_key,        /* Key to search.                       */
      struct path         * p_s_search_path,/* Filled up by this function.          */
      int		  * p_n_pos_in_item,/* returned value, which is the found position in the item */
      int                 * p_n_repeat	    /* Whether schedule occured. */
    ) {
  struct item_head    * p_s_ih;
  int                   n_blk_size;

  /* If searching for directory entry. */
  if ( is_key_of_dir_entry (p_s_key) )
    return  search_by_key(p_s_sb, p_s_key, p_s_search_path, p_n_repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS);

  /* If not searching for directory entry. */

  /* If item is found. */
  if ( search_by_key(p_s_sb, p_s_key, p_s_search_path, p_n_repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS) == ITEM_FOUND )  {

#ifdef REISERFS_CHECK
    if ( ! PATH_PITEM_HEAD(p_s_search_path)->ih_item_len )
      reiserfs_panic(p_s_sb, "PAP-5165: search_for_position_by_key: item length equals zero");
#endif

    *p_n_pos_in_item = 0;
    return POSITION_FOUND;
  }

#ifdef REISERFS_CHECK
  if ( ! PATH_LAST_POSITION(p_s_search_path) )
    reiserfs_panic(p_s_sb, "PAP-5170: search_for_position_by_key: position equals zero");
#endif

  /* Item is not found. Set path to the previous item. */
  p_s_ih = B_N_PITEM_HEAD(PATH_PLAST_BUFFER(p_s_search_path), --PATH_LAST_POSITION(p_s_search_path));
  n_blk_size = p_s_sb->s_blocksize;

  if ( items_of_different_files( &(p_s_ih->ih_key), p_s_key) ) {
    /* there is no any items of file */
    return POSITION_NOT_FOUND;
  }

#ifdef REISERFS_CHECK
  /* found item must be entry (maybe unlinked) with object's stat data or item of the same file */
  if ( ! is_entry_with_stat_data (p_s_ih) && ((is_key_of_indirect_item (p_s_key) && ! is_indirect_item (p_s_ih)) ||
					      (is_key_of_direct_item (p_s_key) && ! is_direct_item(p_s_ih))) ) {
    print_block (PATH_PLAST_BUFFER(p_s_search_path), PRINT_LEAF_ITEMS, 
		 PATH_LAST_POSITION (p_s_search_path) - 2,
		 PATH_LAST_POSITION (p_s_search_path) + 2);
    reiserfs_panic(p_s_sb, "PAP-5190: search_for_position_by_key: found item type %k\ndoes not match to the expected one %k",
		   p_s_ih->ih_key, p_s_key);
  }
#endif

  /* Needed byte is contained in the item pointed to by the path.*/
  if ( I_K_KEY_IN_ITEM(p_s_ih, p_s_key, n_blk_size) )  {
    *p_n_pos_in_item = p_s_key->k_offset - p_s_ih->ih_key.k_offset;
    if ( I_IS_INDIRECT_ITEM(p_s_ih) )
      *p_n_pos_in_item /= n_blk_size;
    return POSITION_FOUND;
  }

  /* Needed byte is not contained in the item pointed to by the path. Set *p_n_pos_in_item out of the
     item. */
  if ( I_IS_INDIRECT_ITEM(p_s_ih) )
    *p_n_pos_in_item = I_UNFM_NUM(p_s_ih);
  else
    *p_n_pos_in_item = p_s_ih->ih_item_len;
  return POSITION_NOT_FOUND;
}


/* Compare given item and item pointed to by the path. */
int comp_items(
      struct item_head  * p_s_ih,
      struct path       * p_s_path
    ) {
  struct buffer_head  * p_s_bh;
  struct item_head    * p_s_path_item;

  /* Last buffer at the path is not in the tree. */
  if ( ! B_IS_IN_TREE(p_s_bh = PATH_PLAST_BUFFER(p_s_path)) )
    return 1;

#ifdef REISERFS_CHECK
    if ( p_s_bh->b_dev == NODEV )
      reiserfs_panic(0, "PAP-5200: comp_items: device is invalid");
#endif

  /* Last path position is invalid. */
  if ( PATH_LAST_POSITION(p_s_path) >= B_NR_ITEMS(p_s_bh) )
    return 1;
  /* Get item at the path. */
  p_s_path_item = PATH_PITEM_HEAD(p_s_path);
  /* Compare keys. */
  if ( comp_keys(&(p_s_path_item->ih_key), &(p_s_ih->ih_key)) )
    return 1;
  /* Compare other items fields. */
  if ( le16_to_cpu (p_s_path_item->ih_free_space) != p_s_ih->ih_free_space ||
       le16_to_cpu (p_s_path_item->ih_item_len) != p_s_ih->ih_item_len ||
       le16_to_cpu (p_s_path_item->ih_item_location) != p_s_ih->ih_item_location )
    return 1;
  /* Items are equal. */
  return 0;
}


/* this treats preserve of conversion an indirect item to a direct one */
static void treat_preserve (struct super_block * s, struct buffer_head * bh, __u32 * block, struct buffer_head * unbh, int * was_unfm_suspected_recipient)
{
  if (!unbh || !is_buffer_suspected_recipient (unbh)) {
    /* unformatted node is not suspected recipient, therefore we have to preserve indirect item,
       containing pointer to it. Do not set was_unfm_suspected_recipient flag for this */
    add_to_preserve (*block, s);
  } else {
    /* unformatted node is suspected recipient, therefore there is no need to preserve neither indirect
       item, that contains corresponding unfm_ptr nor unformatted node itself */
    if (was_unfm_suspected_recipient == 0)
      reiserfs_panic (s, "vs-5282: treat_preserve: can not set \'was unfm suspect recipient\' flag");
    *was_unfm_suspected_recipient = 1;
    unmark_suspected_recipient (s, unbh);
    reiserfs_free_block(s, *block);
    *block = 0;
    mark_buffer_dirty (bh, 1);
  }
  /* in both cases leaves, contents of this unformatted node falled to, are marked as suspected
     recipients already */
}


/*  If the path points to a directory or direct item, calculate mode and the size cut, for balance.
    If the path points to an indirect item, remove some number of its unformatted nodes.
    In case of file truncate calculate whether this item must be deleted/truncated or last
    unformatted node of this item will be converted to a direct item.
    This function returns a determination of what balance mode the calling function should employ. */
char  prepare_for_delete_or_cut(
        struct super_block  * p_s_sb,
        struct path         * p_s_path,
        struct key          * p_s_item_key,
        int                 * p_n_pos_in_item,
        int                 * p_n_removed,      /* Number of unformatted nodes which were removed
                                                    from end of the file. */
        int                 * p_n_cut_size,
        unsigned long         n_new_file_length, /* MAX_KEY_OFFSET in case of delete. */
	int		      preserve_mode,
	int * was_unfm_suspected_recipient
      ) {
  struct item_head    * p_s_ih = PATH_PITEM_HEAD(p_s_path);


#ifdef REISERFS_CHECK
  int n_repeat_counter = 0;
#endif

  /* Stat_data or directory entry item or direct item to be deleted */
  if ( I_IS_STAT_DATA_ITEM(p_s_ih) || I_IS_DIRECTORY_ITEM(p_s_ih) ||
       (I_IS_DIRECT_ITEM(p_s_ih) && (n_new_file_length == MAX_KEY_OFFSET ||
				     n_new_file_length < p_s_ih->ih_key.k_offset))) {

#ifdef REISERFS_CHECK
    if ( n_new_file_length != MAX_KEY_OFFSET && (I_IS_STAT_DATA_ITEM(p_s_ih) || I_IS_DIRECTORY_ITEM(p_s_ih)) )
      reiserfs_panic(p_s_sb, "PAP-5210: prepare_for_delete_or_cut: can not cut entry item or stat data");
#endif

    *p_n_cut_size = -(IH_SIZE + p_s_ih->ih_item_len);
    return M_DELETE;
  }

  /* Direct item. */
  if ( I_IS_DIRECT_ITEM(p_s_ih) ) {
    /* Calculate first position and size for cutting from item. */
    *p_n_cut_size = -(p_s_ih->ih_item_len -
            (*p_n_pos_in_item = n_new_file_length + 1 - p_s_ih->ih_key.k_offset));
    return M_CUT; /* Cut from this item. */
  }

  /* Case of an indirect item. */
  {
  int                   n_unfm_number,    /* Number of the item unformatted nodes. */
                        n_counter,
                        n_repeat,
                        n_retry,        /* Set to one if there is unformatted node buffer in use. */
    			n_blk_size;
  __u32		      * p_n_unfm_pointer; /* Pointer to the unformatted node number. */
  struct item_head      s_ih;           /* Item header. */
  char                  c_mode;           /* Returned mode of the balance. */
  struct buffer_head  * p_s_un_bh;


  n_blk_size = p_s_sb->s_blocksize;

  /* Search for the needed object indirect item until there are no unformatted nodes to be removed. */
  do  {
    /* Copy indirect item header to a temp variable. */
    copy_item_head(&s_ih, PATH_PITEM_HEAD(p_s_path));
    /* Calculate number of unformatted nodes in this item. */
    n_unfm_number = I_UNFM_NUM(&s_ih);

#ifdef REISERFS_CHECK
    if ( ! I_IS_INDIRECT_ITEM(&s_ih) || ! n_unfm_number || *p_n_pos_in_item + 1 !=  n_unfm_number ) {
      reiserfs_panic(p_s_sb, "PAP-5240: prepare_for_delete_or_cut: item must be indirect one (%h)\n\
 or invalid position %d in it", &s_ih, *p_n_pos_in_item);
    }
#endif

    /* Calculate balance mode and position in the item to remove unformatted nodes. */
    if ( n_new_file_length == MAX_KEY_OFFSET || n_new_file_length < s_ih.ih_key.k_offset ) {
      /* indirect item is to be deleted */
      *p_n_pos_in_item = 0;
      *p_n_cut_size = -(IH_SIZE + s_ih.ih_item_len);
      c_mode = M_DELETE;
    } else {
      /* indirect item must be truncated starting from *p_n_pos_in_item-th position */
      *p_n_pos_in_item = (n_new_file_length + n_blk_size - s_ih.ih_key.k_offset ) / n_blk_size;
      
#ifdef REISERFS_CHECK
      if ( *p_n_pos_in_item > n_unfm_number ) 
	reiserfs_panic(p_s_sb, "PAP-5250: prepare_for_delete_or_cut: illegal position in the item");
#endif

      /* Either convert last unformatted node of indirect item to direct item or increase
	 its free space.  */
      if ( *p_n_pos_in_item == n_unfm_number )  {
	*p_n_cut_size = 0; /* Nothing to cut. */
	return M_CONVERT; /* Maybe convert last unformatted node to the direct item. */
      }
      /* Calculate size to cut. */
      *p_n_cut_size = -(s_ih.ih_item_len - *p_n_pos_in_item * UNFM_P_SIZE);
      
      c_mode = M_CUT;     /* Cut from this indirect item. */
    }

  
#ifdef REISERFS_CHECK
    if ( n_unfm_number <= *p_n_pos_in_item ) 
      reiserfs_panic(p_s_sb, "PAP-5260: prepare_for_delete_or_cut: position in the indirect item is too big");
#endif

    /* pointers to be cut */
    n_unfm_number -= *p_n_pos_in_item;
    /* Set pointer to the last unformatted node pointer that is to be cut. */
    p_n_unfm_pointer = (__u32 *)B_I_PITEM(PATH_PLAST_BUFFER(p_s_path),&s_ih) + I_UNFM_NUM(&s_ih)/*vs n_unfm_number*/ - 1 - *p_n_removed;

    /* We go through the unformatted nodes pointers of the indirect item and look for
       the unformatted nodes in the cache. If we found some of them we free it and zero
       corresponding indirect item entry. If some unformatted node has b_count > 1 we must
       not free this unformatted node since it is in use. */
    for ( n_retry = 0, n_counter = *p_n_removed;
	  n_counter < n_unfm_number; n_counter++, p_n_unfm_pointer-- )  {
#ifdef REISERFS_CHECK
      if (p_n_unfm_pointer < (__u32 *)B_I_PITEM(PATH_PLAST_BUFFER(p_s_path),&s_ih) ||
	  p_n_unfm_pointer > (__u32 *)B_I_PITEM(PATH_PLAST_BUFFER(p_s_path),&s_ih) + I_UNFM_NUM(&s_ih) - 1)
	reiserfs_panic (p_s_sb, "vs-5265: prepare_for_delete_or_cut: pointer out of range");
#endif
      if ( ! *p_n_unfm_pointer )  { /* Hole, nothing to remove. */
        if ( ! n_retry )
          (*p_n_removed)++;
        continue;
      }
      /* Search for the unformatted node buffer in the cache */
      n_repeat = CARRY_ON;
      p_s_un_bh = reiserfs_get_hash_table(p_s_sb->s_dev, *p_n_unfm_pointer,
                                                      n_blk_size, &n_repeat);
      if ( n_repeat != CARRY_ON && comp_items(&s_ih, p_s_path) )  {
	/* Current item was shifted from buffer pointed to by the path. */
        brelse(p_s_un_bh);
        break;
      }

      /* Block is in use. */
      if ( p_s_un_bh && p_s_un_bh->b_count != 1 )  {

#ifdef REISERFS_CHECK_ONE_PROCESS
	reiserfs_panic(p_s_sb, "PAP-5270: prepare_for_delete_or_cut: b_count != 1");
#endif

        n_retry = 1;
        brelse(p_s_un_bh);
        continue;
      }
      
      if ( ! n_retry )
        (*p_n_removed)++;
      
#ifdef REISERFS_CHECK
      if ( p_s_un_bh && (*p_n_unfm_pointer != p_s_un_bh->b_blocknr || buffer_locked (p_s_un_bh)))
	reiserfs_panic(p_s_sb, "PAP-5280: prepare_for_delete_or_cut: blocks numbers are different");	
#endif

      if (preserve_mode == PRESERVE_INDIRECT_TO_DIRECT) {
	treat_preserve (p_s_sb, PATH_PLAST_BUFFER(p_s_path), p_n_unfm_pointer, p_s_un_bh, was_unfm_suspected_recipient);
	if (n_unfm_number != 1)
	  reiserfs_panic (p_s_sb, "PAP-5285: prepare_for_delete_or_cut: indirect_to_direct must cut only one pointer (not %d)", n_unfm_number);
      } else {
	/* Free block and buffer in case of removal or cutting. Set unfm ptr in indirect item to 0 */
	unmark_suspected_recipient (p_s_sb, p_s_un_bh);
	reiserfs_free_block(p_s_sb, *p_n_unfm_pointer);
	*p_n_unfm_pointer = 0;
	mark_buffer_dirty (PATH_PLAST_BUFFER(p_s_path), 1);
      }

      /**p_n_unfm_pointer = 0;*/
      bforget (p_s_un_bh);
    }
    /* There is block in use. */
    if ( n_retry )  {

#ifdef REISERFS_CHECK
      if ( *p_n_removed >= n_unfm_number )
	reiserfs_panic(p_s_sb, "PAP-5290: prepare_for_delete_or_cut: unfm pointers to be cut %d, bforgotten buffers %d",
		       n_unfm_number, *p_n_removed);
      if ( !(++n_repeat_counter % 50000) ) {
	reiserfs_warning ("PAP-5300: prepare_for_delete_or_cut: there were %d attempts to get bforget all unformatted node,\
item key %k, new file length %ld",
			  n_repeat_counter, p_s_item_key, n_new_file_length);
      }
#endif

#ifdef __KERNEL__
      current->counter = 0;
      schedule();
#else
      reiserfs_panic (p_s_sb, "vs-5305: prepare_for_delete_or_cut: block in use");
#endif
    }
    /* This loop can be optimized. */
  } while ( *p_n_removed < n_unfm_number &&
          search_for_position_by_key(p_s_sb, p_s_item_key, p_s_path, p_n_pos_in_item, &n_repeat) == POSITION_FOUND )  ;

#ifdef REISERFS_CHECK
  if ( *p_n_removed < n_unfm_number )
    reiserfs_panic(p_s_sb, "PAP-5310: prepare_for_delete_or_cut: indirect item is not found");

  if ( comp_items(&s_ih, p_s_path) ) {
    printk("*p_n_removed = %d n_unfm_number = %d\n",*p_n_removed, n_unfm_number);
    reiserfs_panic(p_s_sb, "PAP-5312: prepare_for_delete_or_cut: item on the path %h does not not match to the saved one %h",
		   PATH_PITEM_HEAD (p_s_path), &s_ih);
  }
#endif

  if (c_mode == M_CUT)
    *p_n_pos_in_item *= UNFM_P_SIZE;
  return c_mode;
  }
}


/* Calculate bytes number which will be deleted or cutted in the balance. */
int calc_deleted_bytes_number(
      struct  tree_balance  * p_s_tb,
      char                    c_mode
    ) {
  int                     n_del_size;
  struct  item_head     * p_s_ih = PATH_PITEM_HEAD(p_s_tb->tb_path);

  if ( I_IS_STAT_DATA_ITEM(p_s_ih) || I_IS_DIRECTORY_ITEM(p_s_ih) )
    return p_s_ih->ih_item_len;
  
  n_del_size = ( c_mode == M_DELETE ) ? p_s_ih->ih_item_len : -p_s_tb->insert_size[0];

  if ( I_IS_INDIRECT_ITEM(p_s_ih) )
    n_del_size = (n_del_size/UNFM_P_SIZE)*
      (PATH_PLAST_BUFFER(p_s_tb->tb_path)->b_size) - p_s_ih->ih_free_space;
  return n_del_size;
}

static void init_tb_struct(
              struct tree_balance * p_s_tb,
	      struct super_block  * p_s_sb,
	      struct path         * p_s_path,
              int                   n_size
            ) {
 memset (p_s_tb,'\0',sizeof(struct tree_balance));
 p_s_tb->tb_sb = p_s_sb;
 p_s_tb->tb_path = p_s_path;
 PATH_OFFSET_PBUFFER(p_s_path, ILLEGAL_PATH_ELEMENT_OFFSET) = NULL;
 PATH_OFFSET_POSITION(p_s_path, ILLEGAL_PATH_ELEMENT_OFFSET) = 0;
 p_s_tb->insert_size[0] = n_size;
}


/* this is called by reiserfs_delete_item works in PRESERVE_DIRECT_TO_INDIRECT mode. tb->tb_path
   points to direct item being deleted. That item must be preserved. unfm is unformatted node direct
   item is copied to, unfm becomes suspected recipient */
static void preserve_direct_item (struct tree_balance * tb, struct buffer_head * unfm)
{
  /* unfm is to become suspected recipient */
  preserve_shifted (tb, &PATH_PLAST_BUFFER (tb->tb_path), PATH_H_PPARENT (tb->tb_path, 0), PATH_H_B_ITEM_ORDER (tb->tb_path, 0), unfm);
}


/* it is called in reiserfs_cut_from_item before call to do_balance, which is going to cut last
   unformatted pointer of indirect item. Direct item is in tree already, and buffers containing it
   are marked as suspected recipients. Last unformatted pointer is preserved by (added to preserve list)
   prepare_for_cut_or_delete and corresponding node is forgotten */
void preserve_indirect_item (struct tree_balance * tb)
{
/*  struct buffer_info bi;
  
  bi.bi_bh = PATH_PLAST_BUFFER (tb->tb_path);
  bi.bi_parent = PATH_H_PPARENT (tb->tb_path, 0);
  bi.bi_position = PATH_H_B_ITEM_ORDER (tb->tb_path, 0);*/
  preserve_shifted (tb, &PATH_PLAST_BUFFER (tb->tb_path), PATH_H_PPARENT (tb->tb_path, 0), PATH_H_B_ITEM_ORDER (tb->tb_path, 0), 0);
}


/* it is called in reiserfs_cut_from_item before call to do_balance if
   it is cutting old entry in reiserfs_rename. Node containing new
   entry is marked as suspected recipient */
void preserve_entry (struct tree_balance * tb)
{
  preserve_indirect_item (tb);
}


/* Delete object item. Returns length of deleted item */
int reiserfs_delete_item(
      struct super_block  * p_s_sb,       /* Pointer to the super block.          */
      struct path         * p_s_path,     /* Path to the deleted item.            */
      int		  * p_n_pos_in_item,
      struct key          * p_s_item_key, /* Key of item to be deleted. */
      struct buffer_head  * p_s_un_bh,    /* NULL or unformatted node pointer.    */
      int                 preserve_mode   /* can be PRESERVE_DIRECT_TO_INDIRECT or NOTHING_SPECIAL */
    ) {
  struct tree_balance   s_del_balance;
  struct item_head      s_ih;
  int                   n_repeat,
                        n_ret_value,
                        n_del_size,
                        n_removed;

#ifdef REISERFS_CHECK
  char                  c_mode;
  int			n_iter = 0;
#endif

  init_tb_struct(&s_del_balance, p_s_sb, p_s_path, 0);
  s_del_balance.preserve_mode = preserve_mode;

  while ( 1 ) {
    n_removed = 0;

#ifdef REISERFS_CHECK
    n_iter++;
    c_mode =
#endif

    prepare_for_delete_or_cut(p_s_sb, p_s_path, p_s_item_key, p_n_pos_in_item, &n_removed, &n_del_size, 
			      MAX_KEY_OFFSET, NOTHING_SPECIAL, 0);

#ifdef REISERFS_CHECK
    if ( c_mode != M_DELETE )
      reiserfs_panic(p_s_sb, "PAP-5320: reiserfs_delete_item: mode must be M_DELETE");
#endif

    copy_item_head(&s_ih, PATH_PITEM_HEAD(p_s_path));
    s_del_balance.insert_size[0] = n_del_size;

#ifdef REISERFS_CHECK
    if ( ( ! is_key_of_stat_data(p_s_item_key) && ! is_key_of_dir_entry (p_s_item_key) &&
	   ! I_K_KEY_IN_ITEM (&s_ih, p_s_item_key, p_s_sb->s_blocksize))  ||
	 keys_of_different_item_types (p_s_item_key, &(s_ih.ih_key))) {
      reiserfs_panic(p_s_sb, "PAP-5325: reiserfs_delete_item: found key (%k) does not match\
to the item being deleted %k on the %d attempts", &(s_ih.ih_key), p_s_item_key, n_iter);
    }

    if ( PATH_LAST_POSITION(s_del_balance.tb_path) >= B_NR_ITEMS(PATH_PLAST_BUFFER(s_del_balance.tb_path)) ) {
      reiserfs_panic(p_s_sb, "PAP-5330: reiserfs_delete_item: invalid item number (%d) on %d attempts, must be < %d. key to be deleted %k", 
		     PATH_LAST_POSITION(s_del_balance.tb_path), n_iter,
		     B_NR_ITEMS(PATH_PLAST_BUFFER(s_del_balance.tb_path)), p_s_item_key);
    }
#endif

    n_ret_value = fix_nodes(M_DELETE, &s_del_balance, 0, NULL);

#ifdef REISERFS_CHECK_ONE_PROCESS
    if ( n_ret_value == PATH_INCORRECT )
      reiserfs_panic(p_s_sb,"PAP-5340: reiserfs_delete_item: illegal returned value");
#endif

    if ( n_ret_value != SCHEDULE_OCCURRED && n_ret_value != PATH_INCORRECT )
      break;
    /* schedule() occured while fix_nodes() worked */
    if ( search_for_position_by_key(p_s_sb, p_s_item_key, p_s_path, p_n_pos_in_item, &n_repeat) == POSITION_NOT_FOUND )
      reiserfs_panic(p_s_sb, "PAP-5350: reiserfs_delete_item: item to delete does not exist");
  }
  if ( n_ret_value == NO_DISK_SPACE ) {
    unfix_nodes(&s_del_balance);
    reiserfs_panic(p_s_sb, "PAP-5360: reiserfs_delete_item: "
		   "reiserfs_delete_item should not ever need new disk space, bug");
    return n_ret_value;
  }
 
  /* Here n_ret_value equals CARRY_ON. */
  n_ret_value = calc_deleted_bytes_number(&s_del_balance, M_DELETE);

  if ( p_s_un_bh )  {
    /* We are deleting direct items in a tail, that we are converting
       into an unformatted node. */

#ifdef REISERFS_CHECK
    if ( ! I_IS_DIRECT_ITEM(&s_ih) || ! buffer_uptodate(p_s_un_bh) || p_s_un_bh->b_count != 1 ) {
      print_buffer_head(p_s_un_bh, "unformatted node");
      reiserfs_panic(p_s_sb,"PAP-5370: reiserfs_delete_item: unformatted node is not uptodate or item is not direct item %h",
		     &s_ih);
    }
#endif

    memcpy(p_s_un_bh->b_data + (s_ih.ih_key.k_offset - 1) % (p_s_sb->s_blocksize),
	   B_I_PITEM(PATH_PLAST_BUFFER(p_s_path), &s_ih), n_ret_value);
#ifdef REISERFS_CHECK
    if ( preserve_mode != PRESERVE_DIRECT_TO_INDIRECT )
      reiserfs_panic(p_s_sb, "PAP-5380: reiserfs_delete_item: "
		     "you need to change the code to check converting_tail before preserving");
#endif /* REISERFS_CHECK */

    /* preserve node containing direct item that is being deleted and mark unformatted node
       (p_s_un_bh) getting contents of direct item as suspected recipient */
    preserve_direct_item (&s_del_balance, p_s_un_bh);
    mark_buffer_dirty(p_s_un_bh, 0);
  }

  /* Perform balancing after all resources will be collected at once. */ 
  do_balance(&s_del_balance, 0, NULL, NULL, M_DELETE, REISERFS_KERNEL_MEM, 0/* zeros number */);

  /* Return deleted body length */ 
  return n_ret_value;
}


/* Summary Of Mechanisms For Handling Collisions Between Processes:

 deletion of the body of the object is performed by iput(), with the
 result that if multiple processes are operating on a file, the
 deletion of the body of the file is deferred until the last process
 that has an open inode performs its iput().

 writes and truncates are protected from collisions by use of
 semaphores.

 creates, linking, and mknod are protected from collisions with other
 processes by making the reiserfs_add_entry() the last step in the
 creation, and then rolling back all changes if there was a collision.
 - Hans
*/

/* Delete all items of an object. */
/* This would work faster if it did not use search by key to access data one node at a time, but instead
   resembled read more by doing read ahead. */
void  reiserfs_delete_object(
        struct inode  * p_s_inode       /* Pointer to the object inode. */
      ) {
  struct key            s_item_key;     /* Key to search for a file item. */  
  unsigned long         n_obj_size;     /* Object size. */
  int                   n_repeat,
                        n_deleted,      /* Number of deleted bytes. */
                        n_pos_in_item,  /* Found position in the item. */
    			n_is_last_item = 1;

  struct super_block *  p_s_sb = p_s_inode->i_sb;
  struct path           s_search_path;  /* Path to the last object item. */
  
  init_path (&s_search_path);

  /* Copy key of the object stat_data. */
  copy_key (&s_item_key, INODE_PKEY(p_s_inode));

  /* Get object size. */
  n_obj_size = p_s_inode->i_size;
  /* Case of a directory. */
  if ( S_ISDIR(p_s_inode->i_mode) ) {

#ifdef REISERFS_CHECK
    /* reiserfs_delete_object is called to delete a directory only for empty directories. */
    if ( n_obj_size != EMPTY_DIR_SIZE )
      reiserfs_panic (p_s_sb, "PAP-5390: reiserfs_delete_object: bad empty directory sdize (%lu)", n_obj_size);
#endif
    /* to jump directly to deletion of stat data */
    n_obj_size = 0;

  } else {
    /* Set key offset to search for the last file byte. */
    s_item_key.k_offset = n_obj_size;
    if ( INODE_OFFSET_IN_DIRECT (p_s_inode, n_obj_size))
      mark_key_direct (&s_item_key);
    else
      mark_key_indirect (&s_item_key);
  }


  /* Delete object body. */
  while ( n_obj_size )  {
    /* Search for the last object item. */
    if ( search_for_position_by_key(p_s_sb, &s_item_key, &s_search_path, &n_pos_in_item, &n_repeat) == POSITION_NOT_FOUND ) {
      struct item_head * p_s_ih = PATH_PITEM_HEAD(&s_search_path);
      
#ifdef REISERFS_CHECK
      if (!n_is_last_item)
	reiserfs_panic (p_s_sb, "PAP-5410: reiserfs_delete_object: item to delete doesn't exist");
#endif
      
      reiserfs_warning ("vs-5420: reiserfs_delete_object: item to delete %k, found item %h, i_size %d\n",
			&s_item_key, p_s_ih, p_s_inode->i_size);

      if ( comp_short_keys(&s_item_key, &(p_s_ih->ih_key)) || I_IS_STAT_DATA_ITEM(p_s_ih) ) {
	/* there is no any items of this file */
	n_obj_size = 0;
	break;
      }
      
      s_item_key.k_offset = n_obj_size = p_s_ih->ih_key.k_offset + I_BYTES_NUMBER(p_s_ih, p_s_sb->s_blocksize) - 1;
      /* set position in item to last its byte/unfm_ptr */
      n_pos_in_item--;
      
      n_is_last_item = 0;
    }

    /* Delete last object item. */
    n_deleted = reiserfs_delete_item(p_s_sb, &s_search_path, &n_pos_in_item, &s_item_key, NULL, NOTHING_SPECIAL);

#ifdef REISERFS_CHECK
    if ( n_deleted <= 0 )
      reiserfs_panic(p_s_sb, "reiser-5420: reiserfs_delete_object: this code needs to be fixed to handle ENOSPC");
    if ( n_deleted > n_obj_size )
      reiserfs_panic (p_s_sb, "PAP-5430: reiserfs_delete_object: reiserfs_delete_item returns too big number");
#endif

    n_obj_size -= n_deleted;

#ifdef REISERFS_CHECK
    if ( n_obj_size && s_item_key.k_offset < n_deleted )
	    reiserfs_panic (p_s_sb, "PAP-5440: reiserfs_delete_object: illegal search key offset");
#endif

    /* Update key to search for the new last object item. */
    s_item_key.k_offset -= n_deleted;
    if ( offset (&s_item_key) < p_s_inode->u.reiserfs_i.i_first_direct_byte )
      mark_key_indirect (&s_item_key);

  } /* while (n_obj_size) */


  /* we do not know whether stat data has been already deleted or not */
  /* Set key to search for the item containing object's stat_data. */
  s_item_key.k_offset = SD_OFFSET;

  /* Search for the object stat_data. */
  if ( search_by_key(p_s_sb, &s_item_key, &s_search_path, &n_repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS) == ITEM_NOT_FOUND )
    pathrelse (&s_search_path);
  else
    /* Delete object stat_data. */
    if ( reiserfs_delete_item(p_s_sb, &s_search_path, &n_pos_in_item, &s_item_key, NULL, NOTHING_SPECIAL) < 0 )
      reiserfs_panic (p_s_sb, "PAP-5455: reiserfs_delete_object: no disk space to delete item");

#ifdef REISERFS_CHECK
  s_item_key.k_offset = MAX_KEY_OFFSET;

  /* Try to find any item of the deleted object. */
  if ( search_by_key (p_s_sb, &s_item_key, &s_search_path, &n_repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS) == ITEM_FOUND )
    reiserfs_panic(p_s_sb,"PAP: 5460: reiserfs_delete_object: there is the item of deleted object");

  PATH_LAST_POSITION(&s_search_path)--;
  if (!items_of_different_files (&(PATH_PITEM_HEAD(&s_search_path)->ih_key), &s_item_key)) {
    print_block (PATH_PLAST_BUFFER (&s_search_path), PRINT_LEAF_ITEMS,
		 PATH_LAST_POSITION(&s_search_path) - 2, PATH_LAST_POSITION(&s_search_path) + 2);
    reiserfs_panic(p_s_sb,"PAP-5470: reiserfs_delete_object: item %h of deleted object (%k) found",
		   PATH_PITEM_HEAD(&s_search_path), INODE_PKEY (p_s_inode));
  }
  decrement_counters_in_path(&s_search_path);

#endif

  p_s_inode->i_size = 0;
}


/*********************** Inode part **************************************/
int increment_i_read_sync_counter(
      struct inode  * p_s_inode
    ) {
  int n_repeat = CARRY_ON;

#ifdef REISERFS_CHECK
  int n_repeat_counter = 0;
#endif

  /* Call schedule while this file is being converted. */
  while ( p_s_inode->u.reiserfs_i.i_is_being_converted )  {
#ifdef __KERNEL__
#ifdef REISERFS_CHECK
    if (p_s_inode->u.reiserfs_i.i_read_sync_counter)
      reiserfs_panic (p_s_inode->i_sb, "PAP-5480: increment_i_read_sync_counter: file is read (synced) already");
    if ( !(++n_repeat_counter % 15000) )
      printk ("increment_i_read_sync_counter: (inode=%lu, pid=%d, counter=%d)\n", p_s_inode->i_ino, current->pid, n_repeat_counter);
#endif
    n_repeat |= SCHEDULE_OCCURRED;
    current->counter = 0;
    schedule();
#else
    reiserfs_panic (p_s_inode->i_sb, "vs-5475: increment_i_read_sync_counter: conversion is in progress");
#endif
  }

  p_s_inode->u.reiserfs_i.i_read_sync_counter++;
  return n_repeat;
}


void  decrement_i_read_sync_counter(
        struct inode  * p_s_inode
      ) {

#ifdef REISERFS_CHECK
  if ( ! p_s_inode->u.reiserfs_i.i_read_sync_counter )
    reiserfs_panic (p_s_inode->i_sb, "PAP-5490: increment_i_read_sync_counter: read_sync_counter is zero");
#endif

  p_s_inode->u.reiserfs_i.i_read_sync_counter--;
}


int lock_inode_to_convert(
      struct inode  * p_s_inode
    ) {
  int n_repeat = CARRY_ON;

#ifdef REISERFS_CHECK
  int n_repeat_counter = 0;
#endif

  /* Call schedule() while there is read from this file. */
  while ( p_s_inode->u.reiserfs_i.i_read_sync_counter ) {
#ifdef __KERNEL__
#ifdef REISERFS_CHECK
    if (p_s_inode->u.reiserfs_i.i_is_being_converted)
      reiserfs_panic (p_s_inode->i_sb, "PAP-5495: lock_inode_to_convert: file is being truncated (or appended) already");
    if ( !(++n_repeat_counter % 15000) )
      printk ("lock_inode_to_convert: (inode=%lu, pid=%d, counter=%d)\n", p_s_inode->i_ino, current->pid, n_repeat_counter);
#endif
    n_repeat |= SCHEDULE_OCCURRED;
    current->counter = 0;
    schedule();
#else
    reiserfs_panic (p_s_inode->i_sb, "vs-5497: lock_inode_to_convert: reading is in progress");
#endif
  }

#ifdef REISERFS_CHECK
  if ( p_s_inode->u.reiserfs_i.i_is_being_converted || p_s_inode->u.reiserfs_i.i_read_sync_counter )
    reiserfs_panic (p_s_inode->i_sb, "PAP-5500: lock_inode_to_convert: illegal case");
#endif

  /* Mark file as ready to convert. */
  p_s_inode->u.reiserfs_i.i_is_being_converted = 1;
  return n_repeat;
}


void  unlock_inode_after_convert(
        struct inode * p_s_inode
      ) {

#ifdef REISERFS_CHECK
  if ( p_s_inode->u.reiserfs_i.i_is_being_converted != 1 ||
                                        p_s_inode->u.reiserfs_i.i_read_sync_counter )
    reiserfs_panic (p_s_inode->i_sb, "PAP-5510: unlock_inode_after_convert: illegal case");
#endif

  /* Read is possible. */
  p_s_inode->u.reiserfs_i.i_is_being_converted = 0;
}

/************ End of the inode part ***************************/






/* Convert an unformatted node to a direct item. 
   Returns number of deleted bytes.  */
static int indirect_to_direct(
      struct inode        * p_s_inode,          /* Pointer to the file inode.                 */
      struct super_block  * p_s_sb,             /* Pointer to the super block.                */
      struct path         * p_s_path,           /* Pointer to the path to the indirect item.  */
      struct key          * p_s_item_key,       /* Key to search for the last file byte.      */
      unsigned long         n_new_file_size,    /* New file size.                             */
      char                * p_c_mode
    ) {
  struct buffer_head  * p_s_unfm_bh;              /* Pointer to the converted unformatted node
                                                    buffer.                                   */
  struct item_head      s_ih;
  unsigned long         n_unfm_number = 0; 	/* Unformatted node block number              */
  int                   n_pos_in_item,
                        n_repeat_or_retval, /* this variable is overloaded to be used for two purposes:
					       tracking whether schedule occured, and for use as
					       a temporary variable */
                        n_block_size = p_s_sb->s_blocksize;

  /* Copy item at the path. */
  copy_item_head(&s_ih, PATH_PITEM_HEAD(p_s_path) );
  /* Don't read while we are converting the unformatted node. */
  if ( (n_repeat_or_retval = lock_inode_to_convert(p_s_inode)) )
    /* Check whether saved item is at the path. */
    n_repeat_or_retval = comp_items(&s_ih,p_s_path);
  if ( n_repeat_or_retval == CARRY_ON )
    /* Calculate last unformatted node number. */
    n_unfm_number = B_I_POS_UNFM_POINTER(PATH_PLAST_BUFFER(p_s_path), &s_ih, I_UNFM_NUM(&s_ih) - 1);

  /* We get the pointer to the unformatted to be converted into a direct item. */
  while ( 1 ) {
    if ( n_repeat_or_retval != CARRY_ON ) {
      /* Search for the indirect item. */
      if ( search_for_position_by_key(p_s_sb, p_s_item_key, p_s_path, &n_pos_in_item, &n_repeat_or_retval) == POSITION_NOT_FOUND )
	reiserfs_panic(p_s_sb, "PAP-5520: indirect_to_direct: item to convert does not exist");
      copy_item_head(&s_ih, PATH_PITEM_HEAD(p_s_path) );
      n_unfm_number = B_I_POS_UNFM_POINTER(PATH_PLAST_BUFFER(p_s_path), &s_ih, I_UNFM_NUM(&s_ih) - 1);
    }
    p_s_unfm_bh = NULL;
    if ( n_unfm_number )  {
      /* Read unformatted node to convert. */
      n_repeat_or_retval = CARRY_ON;
      p_s_unfm_bh = reiserfs_bread(p_s_sb->s_dev, n_unfm_number, p_s_sb->s_blocksize, &n_repeat_or_retval);
      /* Current item was shifted from buffer at the path. */
      if ( n_repeat_or_retval != CARRY_ON && comp_items(&s_ih, p_s_path) )  {
        brelse(p_s_unfm_bh);
        continue;
      }

#if defined(REISERFS_CHECK) && !defined(PACKING_LOCALITY_READ_AHEAD)

      if ( p_s_unfm_bh->b_count != 1 ) {
        reiserfs_panic (p_s_sb, "PAP-5530: indirect_to_direct: (read counter %d, converted %d) converted block (%d) must not be in use (b_count==%d)",
			p_s_inode->u.reiserfs_i.i_read_sync_counter, p_s_inode->u.reiserfs_i.i_is_being_converted, n_unfm_number, p_s_unfm_bh->b_count);
      }

#endif

    }
    break;
  }

  /* Set direct item header to insert. */
  s_ih.ih_key.k_offset += (I_UNFM_NUM (&s_ih) - 1) * n_block_size;
  n_pos_in_item = s_ih.ih_key.k_offset;	/*(s_ih.ih_key.k_offset -= (s_ih.ih_key.k_offset - 1) % n_block_size);*/
  mark_key_direct (&(s_ih.ih_key));
  s_ih.ih_free_space = MAX_US_INT;
  mark_direct_item (&s_ih);
  n_repeat_or_retval = s_ih.ih_item_len = n_new_file_size % n_block_size;
  PATH_LAST_POSITION(p_s_path)++;

  /* Insert new direct item in the tree. This insert must mark nodes getting a new item as suspected recipient */
  if ( reiserfs_insert_item(p_s_sb, p_s_path, &s_ih,
			    ( p_s_unfm_bh ) ? p_s_unfm_bh->b_data : NULL, 0/*zero bytes*/, PRESERVE_INDIRECT_TO_DIRECT) != OK /*< 0*/ ) {
    /* No disk memory. So we can not convert last unformatted node to the direct item.
       In this case we mark that node has just 'n_new_file_size % n_block_size'
       bytes of the file.*/
    struct item_head * p_s_ih;

    if ( search_for_position_by_key(p_s_sb, p_s_item_key, p_s_path, &n_pos_in_item, &n_repeat_or_retval) == POSITION_NOT_FOUND )
      reiserfs_panic(p_s_sb, "PAP-5540: indirect_to_direct: item to convert does not exist");
    n_repeat_or_retval = (p_s_ih = PATH_PITEM_HEAD(p_s_path))->ih_free_space;
    p_s_ih->ih_free_space = n_block_size - n_new_file_size % n_block_size;
    mark_buffer_dirty(PATH_PLAST_BUFFER(p_s_path), 0);

#ifdef REISERFS_CHECK
    if ( n_repeat_or_retval >= p_s_ih->ih_free_space )
      reiserfs_panic (p_s_sb, "PAP-5550: indirect_to_direct: illegal new ih_free_space");
#endif

    n_repeat_or_retval = p_s_ih->ih_free_space - n_repeat_or_retval;
    *p_c_mode = M_SKIP_BALANCING;
    unlock_inode_after_convert(p_s_inode);
    pathrelse(p_s_path);
  }
  else {
    /* We have inserted new direct item and must remove last unformatted node. */
    *p_c_mode = M_CUT;
    /* Set position of its first byte to inode (for read needs) */
    p_s_inode->u.reiserfs_i.i_first_direct_byte = n_pos_in_item;
  }

  brelse(p_s_unfm_bh);
  /* We have inserted new direct item and must remove last unformatted node. */
/*  *p_c_mode = M_CUT;*/
  return n_repeat_or_retval;
}



int maybe_indirect_to_direct(
      struct inode        * p_s_inode,
      struct super_block  * p_s_sb,
      struct path         * p_s_path,
      struct key          * p_s_item_key,
      unsigned long         n_new_file_size,
      char                * p_c_mode
    ) {
  int n_block_size = p_s_sb->s_blocksize;
  int cut_bytes;

  /* We can store tail of the file in an unformatted node. */ 
  if ( dont_have_tails (p_s_sb) ||
       STORE_TAIL_IN_UNFM(n_new_file_size, n_new_file_size % n_block_size, n_block_size) ) {
    /* Change ih->u.ih_free_space in the indirect item defined by path. */
    struct item_head  * p_s_ih = PATH_PITEM_HEAD(p_s_path);
    int                 n_old_free_space = p_s_ih->ih_free_space;

    *p_c_mode = M_SKIP_BALANCING;
    p_s_ih->ih_free_space = n_block_size - n_new_file_size % n_block_size;

    mark_buffer_dirty(PATH_PLAST_BUFFER(p_s_path), 0);
    
#ifdef REISERFS_CHECK
    if ( n_old_free_space >= p_s_ih->ih_free_space )
      reiserfs_panic (p_s_sb, "PAP-5560: maybe_indirect_to_direct: tail is too small");
#endif

    cut_bytes = p_s_ih->ih_free_space - n_old_free_space;
    pathrelse(p_s_path);
    return cut_bytes;
  }
  /* Permorm the conversion to a direct_item. */
  return indirect_to_direct(p_s_inode, p_s_sb, p_s_path, p_s_item_key, n_new_file_size, p_c_mode);
}


/* we did indirect_to_direct conversion. And we have inserted direct item successesfully, but there
   were no disk space to cut unfm pointer being converted. Therefore we have to delete inserted
   direct item */
static void indirect_to_direct_roll_back (struct inode * inode, struct path * path)
{
  struct key tail_key;
  int tail_len;
  int pos_in_item;
  int repeat_or_removed;


  copy_key (&tail_key, INODE_PKEY (inode));
  tail_key.k_offset = inode->i_size + 1;
  mark_key_direct (&tail_key);

  tail_len = tail_key.k_offset % inode->i_sb->s_blocksize - 1;
  while (tail_len) {
    /* look for the last byte of the tail */
    if (search_for_position_by_key (inode->i_sb, &tail_key, path, &pos_in_item, &repeat_or_removed) == POSITION_NOT_FOUND)
      reiserfs_panic (inode->i_sb, "vs-5615: indirect_to_direct_roll_back: found invalid item");
#ifdef REISERFS_CHECK
    if (pos_in_item != PATH_PITEM_HEAD (path)->ih_item_len - 1)
      reiserfs_panic (inode->i_sb, "vs-5616: indirect_to_direct_roll_back: appended bytes found");
#endif
    PATH_LAST_POSITION (path) --;
	
    repeat_or_removed = reiserfs_delete_item (inode->i_sb, path, &pos_in_item, &tail_key, 0, NOTHING_SPECIAL);
#ifdef REISERFS_CHECK
    if (repeat_or_removed <= 0 || repeat_or_removed > tail_len)
      reiserfs_panic (inode->i_sb, "vs-5617: indirect_to_direct_roll_back: there was tail %d bytes, removed item length %d bytes",
		      tail_len, repeat_or_removed);
#endif
    tail_len -= repeat_or_removed;
    tail_key.k_offset -= repeat_or_removed;
  }
  printk ("reiserfs_cut_from_item: indirect_to_direct conversion has been rolled back due to lack of disk space\n");
  inode->u.reiserfs_i.i_first_direct_byte = NO_BYTES_IN_DIRECT_ITEM;
  mark_inode_dirty (inode);
}


/* (Truncate or cut entry) or delete object item. */
int reiserfs_cut_from_item(
      struct inode        * p_s_inode,
      struct super_block  * p_s_sb,
      struct path         * p_s_path,
      int                 * p_n_pos_in_item,
      struct key          * p_s_item_key,
      unsigned long         n_new_file_size,
      int		    preserve_mode	/* can be PRESERVE_RENAMING or NOTHING SPECIAL */
    ) {
    /* Every function which is going to call do_balance must first create a tree_balance structure.
       Then it must fill up this structure by using the init_tb_struct and fix_nodes functions.
       After that we can make tree balancing. */
  struct tree_balance s_cut_balance;
  int                 n_repeat,
                      n_cut_size,        /* Amount to be cut. */
                      n_ret_value = CARRY_ON,
                      n_removed = 0,     /* Number of the removed unformatted nodes. */
  		      n_is_inode_locked = 0;
  char                c_mode;            /* Mode of the balance. */
  int was_unfm_suspected_recipient = 0;


  init_tb_struct(&s_cut_balance, p_s_sb, p_s_path, n_cut_size);
  s_cut_balance.preserve_mode = preserve_mode;

  /* Repeat this loop until we either cut the item without needing to balance, or we fix_nodes without
     schedule occuring */
  while ( 1 ) {
      /* Determine the balance mode, position of the first byte to be cut, and size to be cut.
	 In case of the indirect item free unformatted nodes which are pointed to by
	 the cut pointers. */

    c_mode = prepare_for_delete_or_cut(p_s_sb, p_s_path, p_s_item_key, p_n_pos_in_item, &n_removed, &n_cut_size, n_new_file_size,
				       n_is_inode_locked ? PRESERVE_INDIRECT_TO_DIRECT : NOTHING_SPECIAL, &was_unfm_suspected_recipient);
    if ( c_mode == M_CONVERT )  {
	/* convert last unformatted node to direct item or adjust its ih_free_space */
#ifdef REISERFS_CHECK
      if ( n_ret_value != CARRY_ON )
        reiserfs_panic (p_s_sb, "PAP-5570: reiserfs_cut_from_item: can not convert twice");
#endif

      n_ret_value = maybe_indirect_to_direct (p_s_inode, p_s_sb, p_s_path, p_s_item_key,
					      n_new_file_size, &c_mode);
      /* We have cut all item bytes and must stop. */
      if ( c_mode == M_SKIP_BALANCING )
        break;
      n_is_inode_locked = 1;
      /* So, we have performed the first part of the conversion: inserting the new direct item.
	 Now we are removing the last unformatted node. */
      /* Set key and key length to search for the last unformatted node. */
      mark_key_indirect (p_s_item_key);

      n_new_file_size -= n_new_file_size % p_s_sb->s_blocksize;
      p_s_item_key->k_offset = n_new_file_size + 1;
      if ( search_for_position_by_key(p_s_sb, p_s_item_key, p_s_path, p_n_pos_in_item, &n_repeat) == POSITION_NOT_FOUND ){
	reiserfs_panic(p_s_sb, "PAP-5580: reiserfs_cut_from_item: item %k to be converted not found.", p_s_item_key);
      }
      continue;
    }

    s_cut_balance.insert_size[0] = n_cut_size;

    n_ret_value = fix_nodes(c_mode, &s_cut_balance, *p_n_pos_in_item, NULL);
 
#ifdef REISERFS_CHECK_ONE_PROCESS
    if ( n_ret_value == PATH_INCORRECT )
      reiserfs_panic(p_s_sb, "PAP-5600: reiserfs_cut_from_item: "
		     "illegal returned value");
#endif

    if ( n_ret_value != SCHEDULE_OCCURRED && n_ret_value != PATH_INCORRECT )
      break;

    /* else schedule() occured while fix_nodes() worked */
    if ( search_for_position_by_key(p_s_sb, p_s_item_key, p_s_path, p_n_pos_in_item, &n_repeat) == POSITION_NOT_FOUND )
      reiserfs_panic(p_s_sb, "PAP-5610: reiserfs_cut_from_item: item to delete does not exist");
  } /* while */


  if ( n_ret_value == NO_DISK_SPACE ) {
    if ( n_is_inode_locked ) {
      indirect_to_direct_roll_back (p_s_inode, p_s_path);
      return 0;
    }
    /* preserve list requires disk space even in cut/delete mode, though */
    reiserfs_panic(p_s_sb,"PAP-5620: reiserfs_cut_from_item: no disk space in cut/delete mode");
  }

  if ( c_mode != M_SKIP_BALANCING ) {

#ifdef REISERFS_CHECK
/*    if ( n_ret_value >= calc_deleted_bytes_number(&s_cut_balance, c_mode) )
      reiserfs_panic (p_s_sb, "PAP-5630: reiserfs_cut_from_item: returned value is too big");*/
    if (n_ret_value != CARRY_ON)
      reiserfs_panic (p_s_sb, "PAP-5630: ret_value is other than CARRY_ON");
    if ( c_mode == M_PASTE || c_mode == M_INSERT )
      reiserfs_panic (p_s_sb, "PAP-5640: reiserfs_cut_from_item: illegal mode");
#endif
      /* Calculate number of bytes that need to be cut from the item. */
    n_ret_value = calc_deleted_bytes_number(&s_cut_balance, c_mode) - n_ret_value;

    if ( c_mode == M_DELETE ) {
      struct item_head * p_s_ih = B_N_PITEM_HEAD(PATH_PLAST_BUFFER(s_cut_balance.tb_path), PATH_LAST_POSITION(s_cut_balance.tb_path));

      if ( I_IS_DIRECT_ITEM(p_s_ih) && p_s_ih->ih_key.k_offset % p_s_sb->s_blocksize == 1 ) {

#ifdef REISERFS_CHECK
	if ( p_s_inode->u.reiserfs_i.i_first_direct_byte != p_s_ih->ih_key.k_offset )
	  reiserfs_panic (p_s_sb, "PAP-5650: reiserfs_cut_from_item: illegal first direct byte position");
#endif

	p_s_inode->u.reiserfs_i.i_first_direct_byte = NO_BYTES_IN_DIRECT_ITEM;
      }
    }

    if (n_is_inode_locked) {
      /* we are going to cut last unfm ptr, preserve it first. unfm node block number is on preserve list already */
#ifdef REISERFS_CHECK
      if (!I_IS_INDIRECT_ITEM (PATH_PITEM_HEAD (s_cut_balance.tb_path)))
	reiserfs_panic (p_s_sb, "vs-5652: reiserfs_cut_from_item: item must be indirect");
      if (c_mode == M_DELETE && -(PATH_PITEM_HEAD (s_cut_balance.tb_path)->ih_item_len + IH_SIZE) != s_cut_balance.insert_size[0]) {
	reiserfs_panic (p_s_sb, "vs-5653: reiserfs_cut_from_item: can not complete indirect_to_direct conversion of item %h\n\
(DELETE, insert_size==%d)", PATH_PITEM_HEAD (s_cut_balance.tb_path), s_cut_balance.insert_size[0]);
      }
      if (c_mode == M_CUT && s_cut_balance.insert_size[0] != -UNFM_P_SIZE) {
	reiserfs_panic (p_s_sb, "vs-5654: reiserfs_cut_from_item: can not complete indirect_to_direct conversion of item %h\n\
(CUT, insert_size==%d)", PATH_PITEM_HEAD (s_cut_balance.tb_path), s_cut_balance.insert_size[0]);
      }
#endif
      /* we should not preserve indirect item if unformatted node was marked as suspected recipient */
      if (!was_unfm_suspected_recipient)
	preserve_indirect_item (&s_cut_balance);
    }
    if (preserve_mode == PRESERVE_RENAMING)
      preserve_entry (&s_cut_balance);
    do_balance(&s_cut_balance, *p_n_pos_in_item, NULL, NULL, c_mode, REISERFS_KERNEL_MEM, 0/* zero number */);
    if ( n_is_inode_locked )
      unlock_inode_after_convert(p_s_inode);
  }
  return n_ret_value;
}




/* Truncate file to the new size. */
void  reiserfs_truncate_file(
        struct  inode * p_s_inode       /* Pointer to the file inode. New size
                                            already marked in the inode. */
      ) {
  struct item_head    * p_s_ih;         /* Pointer to an item header. */
  struct key            s_item_key;     /* Key to search for a previous file item. */
  unsigned long         n_file_size,    /* Old file size. */
                        n_new_file_size;/* New file size. */
  int                   n_deleted,      /* Number of deleted or truncated bytes. */
                        n_pos_in_item,  /* Found position in an item. */
  			n_repeat;   
  struct path           s_search_path;  /* Path to the current object item. */

  init_path (&s_search_path);

  if ( ! (S_ISREG(p_s_inode->i_mode) || S_ISDIR(p_s_inode->i_mode) || S_ISLNK(p_s_inode->i_mode)) )
    return;


  /* Copy key of the first object item. */
  copy_key(&s_item_key, INODE_PKEY(p_s_inode));

  /* Get new file size. */
  n_new_file_size = p_s_inode->i_size;

  /* Form key to search for the last file item. */
  s_item_key.k_offset = MAX_KEY_OFFSET; /* We don't know old size of the file. */
  if ( p_s_inode->u.reiserfs_i.i_first_direct_byte != NO_BYTES_IN_DIRECT_ITEM )
    mark_key_direct (&s_item_key);
  else
    mark_key_indirect (&s_item_key);

  if ( search_for_position_by_key(p_s_inode->i_sb, &s_item_key, &s_search_path, &n_pos_in_item, &n_repeat) == POSITION_FOUND )
    reiserfs_panic (p_s_inode->i_sb, "PAP-5660: reiserfs_truncate_file: "
		      "object item has too big offset");
  n_pos_in_item--;

  /* Calculate old size of the file. */
  p_s_ih = PATH_PITEM_HEAD(&s_search_path);
  if ( comp_short_keys (&(p_s_ih->ih_key), INODE_PKEY (p_s_inode)) || I_IS_STAT_DATA_ITEM(p_s_ih) )
    n_file_size = 0;
  else
    n_file_size = p_s_ih->ih_key.k_offset + I_BYTES_NUMBER(p_s_ih,p_s_inode->i_sb->s_blocksize) - 1;

  if ( n_file_size == 0 || n_file_size < n_new_file_size ) {
#ifdef REISERFS_INFO
    printk ("reiserfs_truncate_file: old file size = %lu < new file size = %lu\n", n_file_size, n_new_file_size);
#endif
    pathrelse(&s_search_path);
    return;
  }
  


  /* Update key to search for the last file item. */
  s_item_key.k_offset = n_file_size;

  do  {
    /* Cut or delete file item. */
    n_deleted = reiserfs_cut_from_item(p_s_inode, p_s_inode->i_sb, &s_search_path, &n_pos_in_item, &s_item_key, n_new_file_size, NOTHING_SPECIAL);

#ifdef REISERFS_CHECK
    if ( n_deleted > n_file_size ){
      printk("n_deleted =%d  n_file_size =%lu  s_item_key.k_offset =%d ",n_deleted,
	     n_file_size,  s_item_key.k_offset);
	    reiserfs_panic (p_s_inode->i_sb, "PAP-5670: reiserfs_truncate_file: "
                    "reiserfs_truncate_file returns too big number");
    }
#endif

    /* Change key to search the last file item. */
    if ( (s_item_key.k_offset = (n_file_size -= n_deleted)) < p_s_inode->u.reiserfs_i.i_first_direct_byte )
      mark_key_indirect (&s_item_key);

    /* While there are bytes to truncate and previous file item is presented in the tree. */
  } while ( n_file_size > n_new_file_size &&
	    search_for_position_by_key(p_s_inode->i_sb, &s_item_key, &s_search_path, &n_pos_in_item, &n_repeat) == POSITION_FOUND )  ;

#ifdef REISERFS_CHECK
  if ( n_file_size > n_new_file_size )
    reiserfs_panic (p_s_inode->i_sb, "PAP-5680: reiserfs_truncate_file: object item did not find");
#endif

  p_s_inode->i_mtime = p_s_inode->i_ctime = CURRENT_TIME;
  if_in_ram_update_sd (p_s_inode);

}


/* Paste bytes to the existing item. Returns OK if paste was successful, NO_DISK_SPACE otherwise */
int reiserfs_paste_into_item(
	struct super_block  * p_s_sb,	   	/* Pointer to the supoer block.	*/
	struct path         * p_s_search_path,	/* Path to the pasted item.          */
	int                 * p_n_pos_in_item,	/* Paste position in the item above. */
	struct key          * p_s_key,        	/* Key to search for the needed item.*/
	const char          * p_c_body,       	/* Pointer to the bytes to paste.    */
	int                   n_pasted_size,  	/* Size of pasted bytes.             */
	int                   n_mem_mode,     	/* Copy from KERNEL or USER buffer.  */
	int		      n_zeros_num	/* Number of zeros to be pasted.     */
	) {
    struct tree_balance s_paste_balance;
    int                 n_fix_nodes_res,
      			n_repeat;

    if ( n_pasted_size < 0 )
      reiserfs_panic(p_s_sb, "PAP-5690: reiserfs_paste_into_item: illegal pasted size");


/* reiserfs_warning ("paste start: path item %h, pos in item %d, key %k\n", PATH_PITEM_HEAD (p_s_search_path), *p_n_pos_in_item, p_s_key);*/

    init_tb_struct(&s_paste_balance, p_s_sb, p_s_search_path, n_pasted_size);
    s_paste_balance.preserve_mode = NOTHING_SPECIAL;
    while ( (n_fix_nodes_res = fix_nodes(M_PASTE, &s_paste_balance, *p_n_pos_in_item, NULL)) == SCHEDULE_OCCURRED ||
	    n_fix_nodes_res == PATH_INCORRECT )  {

#ifdef REISERFS_CHECK_ONE_PROCESS
      if ( n_fix_nodes_res == PATH_INCORRECT )
	reiserfs_panic(p_s_sb, "PAP-5700: reiserfs_paste_into_item: illegal returned value");
#endif

      /* schedule() occurred while fix_balance() worked */
      if ( search_for_position_by_key (p_s_sb, p_s_key, p_s_search_path, p_n_pos_in_item, &n_repeat) == POSITION_FOUND ) {
	reiserfs_panic (p_s_sb, "PAP-5710: reiserfs_paste_into_item: entry or pasted byte %k already exists", p_s_key);
      }

/* reiserfs_warning ("paste after research: path item %h, pos in item %d, key %k\n", PATH_PITEM_HEAD (p_s_search_path), *p_n_pos_in_item, p_s_key);*/

#ifdef REISERFS_CHECK
      {
	struct item_head * found_ih = PATH_PITEM_HEAD (p_s_search_path);

	if (I_IS_DIRECT_ITEM (found_ih)) {
	  if (found_ih->ih_key.k_offset + I_BYTES_NUMBER (found_ih, p_s_sb->s_blocksize) !=
	      p_s_key->k_offset ||
	      I_BYTES_NUMBER (found_ih, p_s_sb->s_blocksize) != *p_n_pos_in_item)
	    reiserfs_panic (p_s_sb, "PAP-5720: reiserfs_paste_into_item: found direct item (%h) or position (%d) does not match to key (%k)",
			    found_ih, *p_n_pos_in_item, p_s_key);
	}
	if (I_IS_INDIRECT_ITEM (found_ih)) {
	  if (found_ih->ih_key.k_offset + I_BYTES_NUMBER (found_ih, p_s_sb->s_blocksize) != p_s_key->k_offset || 
	      I_UNFM_NUM (found_ih) != *p_n_pos_in_item ||
	      found_ih->ih_free_space != 0)
	    reiserfs_panic (p_s_sb, "PAP-5730: reiserfs_paste_into_item: found indirect item (%h) or position (%d) does not match to key (%k)",
			    found_ih, *p_n_pos_in_item, p_s_key);
	}
      }
#endif
    }

    /* Perform balancing after all resources are collected by fix_nodes, and accessing
      them will not risk triggering schedule. */
    if ( n_fix_nodes_res == CARRY_ON ) {

      if ( s_paste_balance.insert_size[0] < 0 )
	reiserfs_panic (p_s_sb, "PAP-5740: reiserfs_paste_into_item: insert_size = %d\n",s_paste_balance.insert_size[0]);

/*print_tb (M_PASTE, PATH_LAST_POSITION (p_s_search_path), *p_n_pos_in_item, &s_paste_balance, 0);*/

      do_balance(&s_paste_balance, *p_n_pos_in_item, NULL, p_c_body, M_PASTE, n_mem_mode, n_zeros_num);
      return OK;
    }
    unfix_nodes(&s_paste_balance);
    return NO_DISK_SPACE; /* No disk space. */
}


/* Insert new item into the buffer at the path. Returns OK if item inserted successfully and
   NO_DISK_SPACE otherwise */
int reiserfs_insert_item(
			 struct super_block  * 	p_s_sb,           /* Pointer to the super block.          */
			 struct path         * 	p_s_path,         /* Path to the inserteded item.         */
			 struct item_head    * 	p_s_ih,           /* Pointer to the item header to insert.*/
			 const char          * 	p_c_body,         /* Pointer to the bytes to insert.      */
			 int			n_zeros_num,
			 int			preserve_mode	  /* can be
								     PRESERVE_INDIRECT_TO_DIRECT or
								     NOTHING_SPECIAL. if
								     PRESERVE_INDIRECT_TO_DIRECT,
								     mark buffers new item gets into
								     as suspected recipients */
			 ) {
    struct tree_balance s_ins_balance;
    int                 n_fix_nodes_res,
      			n_repeat;


    init_tb_struct(&s_ins_balance, p_s_sb, p_s_path, IH_SIZE + p_s_ih->ih_item_len);
    s_ins_balance.preserve_mode = preserve_mode;

    if (p_c_body == 0)
      n_zeros_num = p_s_ih->ih_item_len;

    while ( (n_fix_nodes_res = fix_nodes(M_INSERT, &s_ins_balance, 0, p_s_ih)) == SCHEDULE_OCCURRED ||
	    n_fix_nodes_res == PATH_INCORRECT ) {

#ifdef REISERFS_CHECK_ONE_PROCESS
    if ( n_fix_nodes_res == PATH_INCORRECT )
      reiserfs_panic(p_s_sb, "PAP-5750: reiserfs_insert_item: illegal returned value");
#endif

      /* schedule occurred while fix_nodes() worked */
      if ( search_by_key(p_s_sb, (struct key *)p_s_ih, p_s_path, &n_repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS) == ITEM_FOUND )
	reiserfs_panic (p_s_sb, "PAP-5760: reiserfs_insert_item: inserted item exists");
    }
    /* make balancing after all resources will be collected at a time */ 
    if ( n_fix_nodes_res == CARRY_ON ) {
      do_balance (&s_ins_balance, 0, p_s_ih, p_c_body, M_INSERT, REISERFS_KERNEL_MEM, n_zeros_num);
      return OK;/*p_s_ih->ih_item_len;*/
    }

    unfix_nodes(&s_ins_balance);
    return NO_DISK_SPACE; /* No disk space */
}




/*********************** range_read code ***************************************/

/* It is interesting to consider why this code is so
   complicated.... It seems like it ought to be simpler.*/


/*  Get data buffer from cache which contains data (byte or directory record) of some object.
    This data has minimal possible key >= than *p_range_begin and <= than *p_range_end.
    In other words get first buffer contains data with key from range
    [*p_key,*p_range_end].
    If it is not possible (needed buffer is not in the cache) prepare (not uptodate) buffer
    at path from root to the needed buffer. Don't read any blocks.
    Returns:    1) via return value     0 if there is not an needed buffer, 1 otherwise;

                2) via pp_s_buf:        NULL if the needed buffer is not in memory, 
                                        or pointer to the needed buffer, or pointer to the prepared buffer;

                3) via p_n_objectid:    Corresponding object id if pointer above points to an unformatted node, MAX_KEY_OBJECTID otherwise.

                4) Recalculated head of the range in p_s_range_head */

int get_buffer_by_range(
      struct super_block      *	p_s_sb,			/* Super block.								*/
      struct key              *	p_s_range_head,		/* Range begin.								*/
      struct key              *	p_s_range_end,  	/* Range end.								*/
      struct buffer_head     **	pp_s_buf,        	/* Returned value; result buffer.					*/
      unsigned long	      *	p_n_objectid		/* Returned value; corresponding object id if *pp_s_buf points to an
							   unformatted node, MAX_KEY_OBJECTID in other cases.			*/
) {			
  int                   n_res,
    			n_pos_in_buffer,
    			n_repeat,
   			n_item_num,
  			n_pos_in_item;
  struct buffer_head  * p_s_bh;             	/* current buffer                       */
  struct key	        s_min_key,
  		      *	p_s_rkey;
  struct item_head    * p_s_ih,
			s_ih;
  unsigned long		n_unfm_pointer;

#ifdef REISERFS_CHECK
  int			n_repeat_counter = 0;
#endif
  struct path		s_path;

  init_path (&s_path);

repeat:

#ifdef REISERFS_CHECK
  if ( ! (++n_repeat_counter % 10000) ) {
    reiserfs_warning ("PAP-5765: get_buffer_by_range: too many repeats (%d). Range head %k",
		      n_repeat_counter, p_s_range_head);
  }
#endif


  /* Search for the needed buffer in the range. */
  n_res = search_by_key(p_s_sb, p_s_range_head, &s_path, &n_repeat, DISK_LEAF_NODE_LEVEL, DONT_READ_BLOCKS);

  n_pos_in_buffer = PATH_LAST_POSITION(&s_path);

#ifdef REISERFS_CHECK
  if ( ! key_in_buffer (&s_path, p_s_range_head, p_s_sb) )
    reiserfs_panic(p_s_sb, " PAP: 5770: get_buffer_by_range: key is not in the buffer");
#endif

  if ( ! buffer_uptodate(p_s_bh = PATH_PLAST_BUFFER(&s_path)) ) {
    /* We can not get data buffer from range. Prepare buffer from the cache and recalculate right delimiting key. */
    p_s_rkey = get_rkey(&s_path, p_s_sb);

    if ( ! comp_keys(p_s_rkey, &MIN_KEY) ) {

#ifdef REISERFS_CHECK_ONE_PROCESS
      reiserfs_panic(p_s_sb, "PAP-5780: get_buffer_by_range: can not get right delimiting key in case of one process");
#endif

#ifdef REISERFS_CHECK
      if ( n_repeat == CARRY_ON )
	reiserfs_panic(p_s_sb, "PAP-5790: get_buffer_by_range: get_rkey returns KEY_MIN");
#endif

      goto repeat;  /* We can not recalculate right delimiting key to continue search for the buffers in the range.
		       Do that by old one. */
    }
    *pp_s_buf = p_s_bh;


#ifdef REISERFS_CHECK
    if ( ! key_in_buffer(&s_path, p_s_range_head, p_s_sb) )
      reiserfs_panic(p_s_sb, "PAP-5791: get_buffer_by_range: key is not in the path");
    if ( s_path.path_length < FIRST_PATH_ELEMENT_OFFSET )
      reiserfs_panic(p_s_sb, "PAP-5792: get_buffer_by_range: path length is too small");
#endif
    
    copy_key(p_s_range_head, p_s_rkey);
    
    s_path.path_length--;
    decrement_counters_in_path(&s_path);
    *p_n_objectid = MAX_KEY_OBJECTID;
    return 1;
  }

  /* last buffer on the path is uptodate */

#ifdef REISERFS_CHECK
  if ( comp_keys(B_N_PKEY(p_s_bh, 0), p_s_range_head) == 1 || comp_keys(B_PRIGHT_DELIM_KEY(p_s_bh), p_s_range_head ) < 1 ) {
    reiserfs_panic(p_s_sb, "PAP-5795: range head key %k we looked for is not in the buffer", p_s_range_head);
  }
  if ( ! B_IS_ITEMS_LEVEL(p_s_bh) || (n_res == ITEM_NOT_FOUND && ! n_pos_in_buffer) )
    reiserfs_panic(p_s_sb, "PAP-5800: get_buffer_by_range: last buffer on the path is not leaf or returned position is 0");
#endif

  p_s_ih = B_N_PITEM_HEAD(p_s_bh, n_pos_in_buffer);
  n_item_num = B_NR_ITEMS(p_s_bh);
  /* Now we are defining the data which are placed in buffer p_s_bh and has minimal key
     more than or equal to *p_range_begin. */
  if ( n_res == ITEM_FOUND ) { /* Item was found in the tree. */
    n_pos_in_item = 0;
    copy_key(&s_min_key,&(p_s_ih->ih_key));
  }

  else {  /* Item was not found in the tree. */
    /* Calculate min_key which is the minimal key of the byte or directory entry
       more or equal than p_range_begin. */
    n_pos_in_item = MAX_INT;
    /*  Previous item is item of the same object we are looking for. */
    if ( ! comp_short_keys(p_s_range_head, &((p_s_ih - 1)->ih_key)) ) {
      if ( I_IS_DIRECTORY_ITEM(p_s_ih - 1) ) {
	if (p_s_range_head->k_offset <= ((p_s_ih - 1)->ih_key).k_offset) {
	/* Search in the directory item for the entry that has minimal key more or equal than *p_s_range_head. */
/*	bin_search_in_dir_item (p_s_ih - 1, B_I_DEH(p_s_bh, p_s_ih - 1), p_s_range_head, &n_pos_in_item);
	if ( n_pos_in_item < I_ENTRY_COUNT(p_s_ih - 1) ) {*/
	/* Previous item contains needed directory entry. */
	  p_s_ih--;
	  PATH_LAST_POSITION(&s_path)--;
	  n_pos_in_buffer--;
	  copy_key(&s_min_key, &p_s_ih->ih_key);
	  /*s_min_key.k_offset = p_s_ih->ih_key.k_offset;*//*B_I_DEH(p_s_bh, p_s_ih)[n_pos_in_item].deh_offset;*/
	  /*s_min_key.k_uniqueness = p_s_ih->ih_key.k_type;*/
	}
	else
	  /* key *p_s_range_head is greater than last entry in directory item */
	  n_pos_in_item = MAX_INT;
      }
      else
	/* key *p_s_range_head is key of regular file */
	if ( I_K_KEY_IN_ITEM((p_s_ih - 1), p_s_range_head, p_s_bh->b_size) ) {
	  /* Previous item contains needed byte. */
	  p_s_ih--;
	  PATH_LAST_POSITION(&s_path)--;
	  n_pos_in_buffer--;
	  copy_key(&s_min_key, p_s_range_head);
	  n_pos_in_item = p_s_range_head->k_offset - p_s_ih->ih_key.k_offset;
	  if ( I_IS_INDIRECT_ITEM(p_s_ih) )
	    n_pos_in_item /= p_s_bh->b_size;
	}
    }

    if ( n_pos_in_item == MAX_INT )
      /* key we looked for is not in the item */
      if ( n_pos_in_buffer == n_item_num )
	copy_key(&s_min_key, &MIN_KEY);
      else {
 	n_pos_in_item = 0;
	copy_key(&s_min_key, &(p_s_ih->ih_key));
      }
  }

  if ( comp_keys(&s_min_key, p_s_range_end) == 1 ) {
 
/******************************************
    if ( ! key_in_buffer(&s_path, p_s_range_head, p_s_sb) )
      reiserfs_panic(p_s_sb, "PAP-2: get_buffer_by_range: path length is too small");
*********************************************/

    *pp_s_buf = NULL;
    copy_key(p_s_range_head, &s_min_key);
    decrement_counters_in_path(&s_path);
    *p_n_objectid = MAX_KEY_OBJECTID;
    return 0;    /* There is no buffer in the range in the tree. */
  }

  if ( ! comp_keys(&s_min_key, &MIN_KEY) ) {
    /* This leaf buffer is not in the range. */
    p_s_rkey = get_rkey(&s_path, p_s_sb);
    if ( ! comp_keys(p_s_rkey, &MIN_KEY) ) {

#ifdef REISERFS_CHECK_ONE_PROCESS
      reiserfs_panic(p_s_sb, "PAP-5810: get_buffer_by_range: can not get right delimiting key in case of one process");
#endif

#ifdef REISERFS_CHECK
      if ( n_repeat == CARRY_ON )
	reiserfs_panic(p_s_sb, "PAP-5820: get_buffer_by_range: get_rkey returns MIN_KEY");
#endif

      goto repeat; /* We can not recalculate right delimiting key to continue search for the buffer in the range.
		      Do that by old one. */
    }

    if ( ! comp_keys(p_s_rkey, &MAX_KEY) ) {

#ifdef REISERFS_CHECK
      if ( ! key_in_buffer(&s_path, p_s_range_head, p_s_sb) )
	reiserfs_panic(p_s_sb, "PAP-5830: get_buffer_by_range: key_in_buffer returned 0");
#endif

      *pp_s_buf = NULL;
      copy_key(p_s_range_head, &MAX_KEY);
      decrement_counters_in_path(&s_path);
      *p_n_objectid = MAX_KEY_OBJECTID;
      return 0;    /* There is no buffer in the range in the tree. */
    }

    *pp_s_buf = NULL;
    copy_key(p_s_range_head, p_s_rkey); /* Reset range head and continue search for the buffer in the range. */
    decrement_counters_in_path(&s_path);
    *p_n_objectid = MAX_KEY_OBJECTID;
    return 1;
  }


  if ( ! I_IS_INDIRECT_ITEM(p_s_ih) ) { /*  We have direct or directory item which contains byte or
                                          directory record in the range in the buffer p_s_bh. */
    /* Look for the next indirect item in the buffer */
    for ( p_s_ih++, n_pos_in_buffer++; n_pos_in_buffer < n_item_num; n_pos_in_buffer++, p_s_ih++ )
      if ( I_IS_INDIRECT_ITEM(p_s_ih) )
        break;

    if ( n_pos_in_buffer == n_item_num ) {
      /* indirect item was not found. */
      p_s_rkey = get_rkey(&s_path, p_s_sb);
      if ( ! comp_keys(p_s_rkey, &MIN_KEY) ) {

#ifdef REISERFS_CHECK_ONE_PROCESS
	reiserfs_panic(p_s_sb, "PAP-5830: get_buffer_by_range: can not get right delimiting key in case of one process");
#endif

#ifdef REISERFS_CHECK
	if ( n_repeat == CARRY_ON )
	  reiserfs_panic(p_s_sb, "PAP-5840: get_buffer_by_range: get_rkey returns MIN_KEY");
#endif
	
	goto repeat; /* We can not recalculate right delimiting key to continue search for the buffers in the range.
			Do that by old one. */
      }
      
#ifdef REISERFS_CHECK
      if ( ! key_in_buffer(&s_path, p_s_range_head, p_s_sb) )
	reiserfs_panic(p_s_sb, "PAP-5850: get_buffer_by_range: key_in_buffer returned 0");

      if ( comp_keys(B_N_PKEY(p_s_bh, 0), p_s_range_head) == 1 || comp_keys(B_PRIGHT_DELIM_KEY(p_s_bh), p_s_range_head ) < 1 ) {
	reiserfs_panic(p_s_sb, "PAP-5860: get_buffer_by_range: range head key %k is not in the last buffer on the path",
		       p_s_range_head);
      }
#endif

      copy_key(p_s_range_head, p_s_rkey); /* Reset range_head. */
    }
    else /* indirect item was found. */ {

#ifdef REISERFS_CHECK
      if ( comp_keys(B_N_PKEY(p_s_bh, 0), p_s_range_head) == 1 || comp_keys(B_PRIGHT_DELIM_KEY(p_s_bh), p_s_range_head ) < 1 ) {
	reiserfs_panic(p_s_sb, "PAP-5870: get_buffer_by_range: range head key %k is not in the last buffer on the path",
		       p_s_range_head);
      }
#endif

/******************************************
      if ( ! key_in_buffer(&s_path, p_s_range_head, p_s_sb) )
	reiserfs_panic(p_s_sb, "PAP-4: get_buffer_by_range: path length is too small");
*********************************************/

      copy_key(p_s_range_head, &(p_s_ih->ih_key)); /* Reset range head. */
    }

    *pp_s_buf = p_s_bh;

#ifdef REISERFS_CHECK
    if ( s_path.path_length < FIRST_PATH_ELEMENT_OFFSET )
      reiserfs_panic(0, "PAP-5880: get_buffer_by_range: path length is too small (%d)", s_path.path_length);
#endif

    s_path.path_length--;
    decrement_counters_in_path(&s_path);
    *p_n_objectid = MAX_KEY_OBJECTID;
    return 1;
  }

  /* Needed byte is located in an unformatted node. Check whether it is in cache.
      And if not prepare buffer to read it. */
  n_unfm_pointer = B_I_POS_UNFM_POINTER(p_s_bh, p_s_ih, n_pos_in_item);
  if ( ! n_unfm_pointer ) { /* This is a hole (nothing to read). */
    if ( n_pos_in_item + 1 == I_UNFM_NUM(p_s_ih) )
      if ( n_pos_in_buffer + 1 == B_NR_ITEMS(p_s_bh) ) {
	p_s_rkey = get_rkey(&s_path, p_s_sb);

	if ( ! comp_keys(p_s_rkey, &MIN_KEY) )
	  goto repeat;

	copy_key(p_s_range_head, p_s_rkey);
      }
      else
	copy_key(p_s_range_head, &((p_s_ih + 1)->ih_key));
    else {
      copy_key(p_s_range_head, &(p_s_ih->ih_key));
      p_s_range_head->k_offset += (n_pos_in_item + 1)*(p_s_sb->s_blocksize);
    }

    *pp_s_buf = NULL;
    decrement_counters_in_path(&s_path);
    *p_n_objectid = p_s_ih->ih_key.file_id;
    return 1;
  } /* unformatted node pointer contains 0 */


  /* Copy found item header. */
  copy_item_head(&s_ih, p_s_ih);
  /* Get unformatted node buffer. */
  n_repeat = CARRY_ON;
  p_s_bh = reiserfs_getblk(p_s_sb->s_dev, n_unfm_pointer, p_s_sb->s_blocksize, &n_repeat);

  if ( n_repeat != CARRY_ON  && comp_items(&s_ih, &s_path) ) {

#ifdef REISERFS_CHECK_ONE_PROCESS
    reiserfs_panic(p_s_sb, "PAP-5890: get_buffer_by_range: item in the path is changed in case of one process");
#endif

    brelse(p_s_bh);
    goto repeat;
  }

#ifdef REISERFS_CHECK
  if ( comp_items(&s_ih, &s_path) )
    reiserfs_panic(p_s_sb, "PAP-5900: get_buffer_by_range: items must be equal");
#endif

  if ( n_pos_in_item + 1 == I_UNFM_NUM(p_s_ih) )
    if ( n_pos_in_buffer + 1 == B_NR_ITEMS(PATH_PLAST_BUFFER(&s_path)) ) {
      p_s_rkey = get_rkey(&s_path, p_s_sb);

      if ( ! comp_keys(p_s_rkey, &MIN_KEY) ) {
	brelse(p_s_bh);
	goto repeat;
      }

      copy_key(p_s_range_head, p_s_rkey);
    }
    else
     copy_key(p_s_range_head, &((p_s_ih + 1)->ih_key));
  else {
    copy_key(p_s_range_head, &(p_s_ih->ih_key));
    p_s_range_head->k_offset += (n_pos_in_item + 1)*(p_s_sb->s_blocksize);
  }

  *pp_s_buf = p_s_bh;
  decrement_counters_in_path(&s_path);
  *p_n_objectid = s_ih.ih_key.file_id;
  return 1; 
}


int get_buffers_from_range(					/*  Returns length of the array of calculated
								    buffers which is less or equal than
								    max_nr_buffers_to_return.         			*/
      struct  super_block     *	p_s_sb,				/*  Pointer to the super block.				*/
      struct  key	      *	p_s_range_start,		/*  Minimal range key.                            	*/
      struct  key	      * p_s_range_end,			/*  Maximal range key.                    		*/
      struct  buffer_head    **	p_s_range_buffers,		/*  Returned array of pointers to buffer headers.
								    Must be allocated in calling function.		*/
      int     			n_max_nr_buffers_to_return	/*  Length of the allocated array.                	*/
    ) {
  struct key          * p_s_cur_key = p_s_range_start;
  struct buffer_head  * p_s_res_buffer;
  int			n_array_length = 0;
  unsigned long		n_objectid;

#ifdef REISERFS_CHECK
  if ( comp_keys(p_s_range_start, p_s_range_end) == 1 ||
       ! comp_keys(p_s_range_start, &MIN_KEY) || ! comp_keys(p_s_range_start, &MAX_KEY) ||
       ! comp_keys(p_s_range_end, &MIN_KEY) || ! comp_keys(p_s_range_end, &MAX_KEY) )
    reiserfs_panic(p_s_sb, "PAP-5890: get_buffers_from_range: illegal range");
#endif

  /* While p_s_cur_key is in the range. */
  while ( comp_keys(p_s_cur_key, p_s_range_end) != 1 )  {
    /* Calculate next buffer from range. */
    if ( ! get_buffer_by_range(p_s_sb, p_s_cur_key, p_s_range_end, &p_s_res_buffer, &n_objectid) )
      break; /* There are not more buffers in the range. */
    p_s_range_buffers[n_array_length++] = p_s_res_buffer;
    if ( n_array_length == n_max_nr_buffers_to_return )
      break;
  }
  return n_array_length;
}


#ifndef REISERFS_FSCK

/*inline int is_left_mergeable (struct item_head * ih, unsigned long bsize)*/
inline int is_left_mergeable (struct key * key, unsigned long bsize)
{
  if (is_key_of_direct_item (key))
    return (key->k_offset % bsize != 1);

  if (is_key_of_indirect_item (key))
    return (key->k_offset != 1);

  /* stat data item and directory entry item are solid */
  return 0;
}

#else

int are_items_mergeable (struct item_head * left, struct item_head * right, int bsize)
{
  if (comp_keys (&left->ih_key, &right->ih_key) != -1) {
    reiserfs_panic (0, "vs-: are_items_mergeable: left %k, right %k", &(left->ih_key), &(right->ih_key));
  }

  if (items_of_different_files (&left->ih_key, &right->ih_key))
    return 0;

  if (I_IS_DIRECTORY_ITEM (left)) {
    return 0;
  }

  if ((I_IS_DIRECT_ITEM (left) && I_IS_DIRECT_ITEM (right)) || 
      (I_IS_INDIRECT_ITEM (left) && I_IS_INDIRECT_ITEM (right)))
    return (left->ih_key.k_offset + I_BYTES_NUMBER (left, bsize) == right->ih_key.k_offset) ? 1 : 0;

  return 0;
}

/* get left neighbor of the leaf node */
static struct buffer_head * get_left_neighbor (struct super_block * s, struct path * path)
{
  struct key key;
  struct path path_to_left_neighbor;
  struct buffer_head * bh;
  int repeat;

  copy_key (&key, B_N_PKEY (PATH_PLAST_BUFFER (path), 0));
  decrement_key (&key);

  search_by_key (s, &key, &path_to_left_neighbor, &repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS);
  if (PATH_LAST_POSITION (&path_to_left_neighbor) == 0) {
    pathrelse (&path_to_left_neighbor);
    return 0;
  }
  bh = PATH_PLAST_BUFFER (&path_to_left_neighbor);
  bh->b_count ++;
  pathrelse (&path_to_left_neighbor);
  return bh;
}

static struct buffer_head * get_right_neighbor (struct super_block * s, struct path * path)
{
  struct key key;
  struct key * rkey;
  int repeat;
  struct path path_to_right_neighbor;
  struct buffer_head * bh;

  rkey = get_rkey (path, s);
  if (comp_keys (rkey, &MIN_KEY) == 0)
    reiserfs_panic (s, "vs-: get_right_neighbor: get_rkey returned min key (path has changed)");
  copy_key (&key, rkey);

  search_by_key (s, &key, &path_to_right_neighbor, &repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS);
  if (PATH_PLAST_BUFFER (&path_to_right_neighbor) == PATH_PLAST_BUFFER (path)) {
    pathrelse (&path_to_right_neighbor);
    return 0;
  }
  bh = PATH_PLAST_BUFFER (&path_to_right_neighbor);
  bh->b_count ++;
  pathrelse (&path_to_right_neighbor);
  return bh;
}


int is_left_mergeable (struct super_block * s, struct path * path)
{
  struct item_head * right;
  struct buffer_head * bh;
  int retval;
  
  right = B_N_PITEM_HEAD (PATH_PLAST_BUFFER (path), 0);

  bh = get_left_neighbor (s, path);
  if (bh == 0) {
    return 0;
  }
  retval = are_items_mergeable (B_N_PITEM_HEAD (bh, B_NR_ITEMS (bh) - 1), right, bh->b_size);
  brelse (bh);
  return retval;
}


int is_right_mergeable (struct super_block * s, struct path * path)
{
  struct item_head * left;
  struct buffer_head * bh;
  int retval;
  
  left = B_N_PITEM_HEAD (PATH_PLAST_BUFFER (path), B_NR_ITEMS (PATH_PLAST_BUFFER (path)) - 1);

  bh = get_right_neighbor (s, path);
  if (bh == 0) {
    return 0;
  }
  retval = are_items_mergeable (left, B_N_PITEM_HEAD (bh, 0), bh->b_size);
  brelse (bh);
  return retval;
}


#endif /* REISERFS_FSCK */
