/*
 * Copyright 1996, 1997, 1998, 1999 Hans Reiser
 */

/* mkreiserfs (0.91) is very simple. It supports only 1, 4 and 8K blocks. It skip first 8K of
   device, and then write super block, two copies of bitmap (size of bitmap is calculated based on
   file system size), and root block. Mkreiserfs assumes, that device does not have unreadable
   blocks, and it occupies first blocks for super, bitmap and root blocks.
   mkreiserfs reserves REISERFS_LOG_SIZE blocks at the end of file system for logging needs.
*/


#include "../include/nokernel.h"
#include <sys/ioctl.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>



void check_and_free_mem (void);
void die (char * fmt, ...);
int is_mounted (char * device_name);





#define BLOCK_SIZE 1024
#define DEFAULT_BLOCKSIZE 4096
#define MIN_BLOCK_AMOUNT (100+REISERFS_LOG_SIZE)


struct buffer_head * g_sb_bh;
struct buffer_head * g_bitmap_bh;
struct buffer_head * g_rb_bh;


int g_i_block_size = DEFAULT_BLOCKSIZE;
int g_i_block_amount;

/* Given a file descriptor and an offset, check whether the offset is
   a valid offset for the file - return 0 if it isn't valid or 1 if it
   is */
int valid_offset( int fd, int offset )
{
  char ch;

  if (lseek (fd, offset, 0) < 0)
    return 0;

  if (read (fd, &ch, 1) < 1)
    return 0;

  return 1;
}



/* calculates number of blocks on device 
 */
unsigned long count_blocks (char * filename, int blocksize)
{
  unsigned long high, low;
  int fd;
  long size;

  fd = open (filename, O_RDONLY);
  if (fd < 0)
    die ("count_blocks: open failed (%s)", strerror (errno));

  if (ioctl (fd, BLKGETSIZE, &size) >= 0) {
    close (fd);
    return  size / (blocksize / 512);
  }


  low = 0;
  for( high = 1; valid_offset (fd, high); high *= 2 )
    low = high;
  while (low < high - 1) {
    const int mid = ( low + high ) / 2;
      
    if (valid_offset (fd, mid))
      low = mid;
    else
      high = mid;
  }
  valid_offset (fd, 0);
  close (fd);
  
  return (low + 1) / (blocksize);
}



/* form super block */
void make_super_block (int dev)
{
  struct reiserfs_super_block * sb;
  unsigned long * oids;


  if (SB_SIZE > g_i_block_size)
    die ("mkreiserfs: blocksize (%d) too small", g_i_block_size);
    

  /* get memory for super block */
/*  g_cp_super_block = getmem (g_i_block_size);*/
  g_sb_bh = getblk (dev, (REISERFS_FIRST_BLOCK * BLOCK_SIZE) / g_i_block_size, g_i_block_size);

/*  sb = (struct reiserfs_super_block *)g_cp_super_block;*/
  sb = (struct reiserfs_super_block *)g_sb_bh->b_data;
  sb->s_blocksize = g_i_block_size;   		/* block size in bytes */
  sb->s_block_count = g_i_block_amount;	/* how many block reiserfs must occupy */
  sb->s_state = REISERFS_VALID_FS;
  sb->s_tree_height = 2;
  sb->s_bmap_nr = g_i_block_amount / (g_i_block_size * 8) + ((g_i_block_amount % (g_i_block_size * 8)) ? 1 : 0);
  memcpy (sb->s_magic, REISERFS_SUPER_MAGIC, sizeof (REISERFS_SUPER_MAGIC));
/*  sb->s_log_start = sb->s_block_count - REISERFS_LOG_SIZE;*/
    

  /* initialize object map */
  oids = (unsigned long *)(sb + 1);
  oids[0] = 1;
  oids[1] = REISERFS_ROOT_OBJECTID + 1;	/* objectids > REISERFS_ROOT_OBJECTID are free */
  sb->s_oid_cursize = 2;

  /* max size must be even */
  sb->s_oid_maxsize = (g_i_block_size - SB_SIZE) / sizeof(unsigned long) / 2 * 2;

  mark_buffer_dirty (g_sb_bh, 0);
  mark_buffer_uptodate (g_sb_bh, 0);
  return;

}


/* this only sets few first bits in bitmap block. Fills not initialized
   fields of super block (root block and bitmap block numbers)
   */
void make_bitmap (void)
{
  struct reiserfs_super_block * sb = (struct reiserfs_super_block *)g_sb_bh->b_data;
  int i, j;
  
  /* get buffer for bitmap block */
  g_bitmap_bh = getblk (g_sb_bh->b_dev, g_sb_bh->b_blocknr + 1, g_sb_bh->b_size);
  
  /* mark, that first 8K of device is busy */
  for (i = 0; i < REISERFS_FIRST_BLOCK / (g_i_block_size / BLOCK_SIZE); i ++)
    set_bit (i, g_bitmap_bh->b_data);

  /* mark that super block is busy */
  set_bit (i++, g_bitmap_bh->b_data);

  /* mark bitmap blocks as busy. We have two bitmaps in current version */
  for (j = 0; j < sb->s_bmap_nr * 2; j++)
    set_bit (i ++, g_bitmap_bh->b_data);
  
  /* and tree root is busy */
  set_bit (i, g_bitmap_bh->b_data);
  sb->s_root_block = i;
  sb->s_free_blocks = sb->s_block_count - i - 1 - REISERFS_LOG_SIZE;
  mark_buffer_dirty (g_bitmap_bh, 0);
  mark_buffer_uptodate (g_bitmap_bh, 0);

  mark_buffer_dirty (g_sb_bh, 0);
  return;
}


/* form the root block of the tree (the block head, the item head, the
   root directory) */
void make_root_block ()
{
  struct reiserfs_super_block * sb = (struct reiserfs_super_block *)g_sb_bh->b_data;
  char * rb;
  struct block_head * blkh;
  struct item_head * ih;

  struct stat_data * sd;
  struct key maxkey = {{3,0x3fffffff}, 0xffffffff, 0xffffffff, 0xffffffff};

  /* get memory for root block */
/*  g_cp_root_block = getmem (g_i_block_size);*/
  g_rb_bh = getblk (g_sb_bh->b_dev, g_sb_bh->b_blocknr + sb->s_bmap_nr * 2 + 1, g_sb_bh->b_size);
  rb = g_rb_bh->b_data;

  /* block head */
  blkh = (struct block_head *)rb;
  blkh->blk_level = DISK_LEAF_NODE_LEVEL;
  blkh->blk_nr_item = 0;
  blkh->blk_free_space = sb->s_blocksize - BLKH_SIZE;
  memcpy (&blkh->blk_right_delim_key, &maxkey, KEY_SIZE);


  /* first item is stat data item of root directory */
  ih = (struct item_head *)(blkh + 1);
  ih->ih_key.group_locale.major = REISERFS_ROOT_OBJECTID;
  ih->ih_key.group_locale.minor = 0;
  ih->ih_key.name_hash = 0;
  ih->ih_key.file_id = REISERFS_ROOT_OBJECTID;
  ih->ih_key.k_offset = 0;
  mark_stat_data (ih);
  ih->ih_free_space = MAX_US_INT;
  ih->ih_item_len = SB_SIZE;
  ih->ih_item_location = sb->s_blocksize - ih->ih_item_len;

  /* fill stat data */
  sd = (struct stat_data *)(rb + ih->ih_item_location);
  sd->sd_mode = S_IFDIR + 0755;
  sd->sd_nlink = 3;
  sd->sd_uid = 0;	/*??*/
  sd->sd_gid = 0;	/*??*/
  sd->sd_size = EMPTY_DIR_SIZE;
  sd->sd_atime = sd->sd_ctime = sd->sd_mtime = time (NULL);
  sd->u.parent_dir_id = REISERFS_ROOT_PARENT_OBJECTID;

  blkh->blk_nr_item ++;
  blkh->blk_free_space -= (IH_SIZE + ih->ih_item_len);

  
  mark_buffer_dirty (g_rb_bh, 0);
  mark_buffer_uptodate (g_rb_bh, 0);
  return;
}


/*
 *  write the super block, the bitmap blocks and the root of the tree
 */
void write_super_and_root_blocks ()
{
  struct reiserfs_super_block * sb = (struct reiserfs_super_block *)g_sb_bh->b_data;
  int i;
/*  int log_block = sb->s_log_start;*/

  /* super block */
  bwrite (g_sb_bh);

  /* bitmap blocks */
  for (i = 0; i < sb->s_bmap_nr; i ++) {
    if (i != 0) {
      g_bitmap_bh->b_blocknr ++;
      memset (g_bitmap_bh->b_data, 0, g_bitmap_bh->b_size);
    }
    if (i == sb->s_bmap_nr - 1) {
      int j;

      /* fill unused part of last bitmap block with 1s */
      for (j = sb->s_block_count % (sb->s_blocksize * 8); j < sb->s_blocksize * 8; j ++) {
	set_bit (j, g_bitmap_bh->b_data);
      }
    }

    /* mark log blocks busy */
/*
    while (log_block < sb->s_block_count && (log_block >= i * sb->s_blocksize * 8 &&
					     log_block < (i + 1) * sb->s_blocksize * 8)) {
      set_bit (log_block % (sb->s_blocksize * 8), g_bitmap_bh->b_data);
      log_block ++;
    }
*/

    /* write true bitmap */
    mark_buffer_dirty (g_bitmap_bh, 0);
    bwrite (g_bitmap_bh);

    /* write cautious bitmap */
    g_bitmap_bh->b_blocknr += sb->s_bmap_nr;
    mark_buffer_dirty (g_bitmap_bh, 0);
    bwrite (g_bitmap_bh);
    g_bitmap_bh->b_blocknr -= sb->s_bmap_nr;
  }
/*
  if (log_block != sb->s_block_count)
    die ("mkreiserfs: last log block and block amount do not match");*/

  /* root block */
  bwrite (g_rb_bh);


  brelse (g_rb_bh);
  brelse (g_bitmap_bh);
  brelse (g_sb_bh);
}


void report ()
{
  struct reiserfs_super_block * sb = (struct reiserfs_super_block *)g_sb_bh->b_data;
  unsigned int i;

  printf ("Block size %d bytes\n", sb->s_blocksize);
  printf ("Block count %d\n", sb->s_block_count);
  printf ("Used blocks (%d): (", sb->s_block_count - sb->s_free_blocks);

  i = 0;
  while (test_bit (i, g_bitmap_bh->b_data)) {
    if (i * sb->s_blocksize < REISERFS_FIRST_BLOCK * BLOCK_SIZE)
      printf ("%d ", i);
    else if (i * sb->s_blocksize == REISERFS_FIRST_BLOCK * BLOCK_SIZE)
      printf ("- skipped), %d - super block, (", i);
    else if (i != sb->s_root_block)
      printf ("%d ", i);
    else
      printf ("- bitmaps), %d - tree root\n", i);
    i ++;
  }
/*  printf ("Reserved %d\n", sb->s_block_count - sb->s_log_start);*/

  fflush (stdout);
}




int main (int argc, char **argv)
{
  char *tmp;
  int dev;
  struct stat statbuf;
  char * device_name;
  char c;

  reiserfs_warning ("\n\n<-----------MKREISERFS, version 0.91a, 1999----------->\n\n");

  if (argc < 2)
    die ("Usage: %s [-k size in kilobytes] /dev/name [blocks]\n\n", argv[0]);

  init_buffer_cache ();

  while ( ( c = getopt( argc, argv, "k:" ) ) != EOF )
    switch( c )
      {
      case 'k' :                  /* -k n - where n is 4|8 */
	g_i_block_size = (int) strtol (optarg, &tmp, 0);
	if ( *tmp || ( g_i_block_size != 1 && g_i_block_size != 4 && g_i_block_size != 8 ))
	  die ("bad block size : %s\n", optarg);
	g_i_block_size *= BLOCK_SIZE;
	break;

      default :
	die ("Usage: %s [-k size in kilobytes] /dev/name [blocks]\n\n", argv[0]);
      }
  device_name = argv [optind];
  

  /* get block number for file system */
  if (optind == argc - 2) {
    g_i_block_amount = (int) strtol (argv[optind + 1], &tmp, 0);
    if (*tmp == 0) {    /* The string is integer */
      if (g_i_block_amount > count_blocks (device_name, g_i_block_size) || g_i_block_amount < MIN_BLOCK_AMOUNT)
	die ("mkreiserfs: specified block amount (%d) is too big or too small", g_i_block_amount);
    }
  } else 
    if (optind == argc - 1) {
      /* number of blocks is not specified */
      g_i_block_amount = count_blocks (device_name, g_i_block_size);
      tmp = "";
    }
    else
      die ("Usage: %s [-k size in kilobytes] /dev/name [blocks]\n\n", argv[0]);

  g_i_block_amount = g_i_block_amount / 8 * 8;


  if (*tmp || g_i_block_amount % 8)
    /* block amount specified is not a valid integer */
    die ("mkreiserfs: bad block count : %s\n", argv[optind + 1]);

  if (is_mounted (device_name))
    die ("mkreiserfs: '%s' contains a mounted file system\n", device_name);

  /* open_device will die if it could not open device */
  dev = open (device_name, O_RDWR);
  if (dev == -1)
    die ("mkreiserfs: can not open '%s': %s", device_name, strerror (errno));
  
  if (fstat (dev, &statbuf) < 0)
    die ("mkreiserfs: unable to stat %s", device_name);
  
  if (!S_ISBLK (statbuf.st_mode))
    die ("mkreiserfs: '%s (%o)' is not a block device", device_name, statbuf.st_mode);
  else        /* Ignore any 'full' fixed disk devices */
    if ( statbuf.st_rdev == 0x0300 || statbuf.st_rdev == 0x0340 
	 || statbuf.st_rdev == 0x0400 || statbuf.st_rdev == 0x0410
	 || statbuf.st_rdev == 0x0420 || statbuf.st_rdev == 0x0430
	 || statbuf.st_rdev == 0x0d00 || statbuf.st_rdev == 0x0d40 )
      /* ???? */
      die ("mkreiserfs: will not try to make filesystem on '%s'", device_name);

  
  /* these fill buffers (super block, first bitmap, root block) with
     reiserfs structures */
  make_super_block (dev);
  make_bitmap ();
  make_root_block ();
  
  report ();

  printf ("ATTENTION: ALL DATA WILL BE LOST ON '%s'! (y/n)", device_name);
  c = getchar ();
  if (c != 'y' && c != 'Y')
    die ("mkreiserfs: Disk was not formatted");

  write_super_and_root_blocks ();

  check_and_free_mem ();


  return 0;

}
